# Copyright (c) 2021 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

from mautrix.api import Method, Path
from mautrix.types import RoomID, RoomTagAccountDataEventContent, RoomTagInfo, Serializable

from ..base import BaseClientAPI


class RoomTaggingMethods(BaseClientAPI):
    """
    Methods in section 13.18 Room Tagging of the spec. These methods are used for organizing rooms
    into tags for the local user.

    See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#room-tagging>`__"""

    async def get_room_tags(self, room_id: RoomID) -> RoomTagAccountDataEventContent:
        """
        Get all tags for a specific room. This is equivalent to getting the ``m.tag`` account data
        event for the room.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#get-matrix-client-r0-user-userid-rooms-roomid-tags>`__

        Args:
            room_id: The room ID to get tags from.

        Returns:
            The m.tag account data event.
        """
        resp = await self.api.request(Method.GET, Path.user[self.mxid].rooms[room_id].tags)
        return RoomTagAccountDataEventContent.deserialize(resp)

    async def get_room_tag(self, room_id: RoomID, tag: str) -> Optional[RoomTagInfo]:
        """
        Get the info of a specific tag for a room.

        Args:
            room_id: The room to get the tag from.
            tag: The tag to get.

        Returns:
            The info about the tag, or ``None`` if the room does not have the specified tag.
        """
        resp = await self.get_room_tags(room_id)
        try:
            return resp.tags[tag]
        except KeyError:
            return None

    async def set_room_tag(self, room_id: RoomID, tag: str, info: Optional[RoomTagInfo] = None) -> None:
        """
        Add or update a tag for a specific room.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#put-matrix-client-r0-user-userid-rooms-roomid-tags-tag>`__

        Args:
            room_id: The room ID to add the tag to.
            tag: The tag to add.
            info: Optionally, information like ordering within the tag.
        """
        await self.api.request(Method.PUT, Path.user[self.mxid].rooms[room_id].tags[tag],
                               content=(info.serialize() if isinstance(info, Serializable)
                                        else (info or {})))

    async def remove_room_tag(self, room_id: RoomID, tag: str) -> None:
        """
        Remove a tag from a specific room.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#delete-matrix-client-r0-user-userid-rooms-roomid-tags-tag>`__

        Args:
            room_id: The room ID to remove the tag from.
            tag: The tag to remove.
        """
        await self.api.request(Method.DELETE, Path.user[self.mxid].rooms[room_id].tags[tag])



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, Optional, Union
import asyncio
import uuid

from mautrix.types import (RoomKeyRequestEventContent, KeyRequestAction, RequestedKeyInfo,
                           EncryptionAlgorithm, RoomID, IdentityKey, SessionID, UserID, DeviceID,
                           EventType, ForwardedRoomKeyEventContent)

from .base import BaseOlmMachine
from .types import DecryptedOlmEvent
from .sessions import InboundGroupSession


class KeyRequestingMachine(BaseOlmMachine):
    async def request_room_key(self, room_id: RoomID, sender_key: IdentityKey,
                               session_id: SessionID, from_devices: Dict[UserID, List[DeviceID]],
                               timeout: Optional[Union[int, float]] = None) -> bool:
        """
        Request keys for a Megolm group session from other devices.

        Once the keys are received, or if this task is cancelled (via the ``timeout`` parameter),
        a cancel request event is sent to the remaining devices.

        Args:
            room_id: The room where the session is used.
            sender_key: The key of the user who created the session.
            session_id: The ID of the session.
            from_devices: A dict from user ID to list of device IDs whom to ask for the keys.
            timeout: The maximum number of seconds to wait for the keys.

        Returns:
            ``True`` if the keys were received and are now in the crypto store, ``False`` otherwise.
        """
        request_id = str(uuid.uuid1())
        request = RoomKeyRequestEventContent(action=KeyRequestAction.REQUEST,
                                             body=RequestedKeyInfo(
                                                 algorithm=EncryptionAlgorithm.MEGOLM_V1,
                                                 room_id=room_id,
                                                 sender_key=sender_key,
                                                 session_id=session_id,
                                             ),
                                             request_id=request_id,
                                             requesting_device_id=self.client.device_id)

        fut = self.loop.create_future()
        self._key_request_waiters[session_id] = fut
        await self.client.send_to_device(EventType.ROOM_KEY_REQUEST,
                                         {user_id: {device_id: request
                                                    for device_id in devices}
                                          for user_id, devices in from_devices.items()})
        got_keys = False
        try:
            user_id, device_id = await asyncio.wait_for(fut, timeout=timeout)
            got_keys = True
            try:
                del from_devices[user_id][device_id]
                if len(from_devices[user_id]) == 0:
                    del from_devices[user_id]
            except KeyError:
                pass
        except (asyncio.CancelledError, asyncio.TimeoutError):
            pass
        del self._key_request_waiters[session_id]
        if len(from_devices) > 0:
            cancel = RoomKeyRequestEventContent(action=KeyRequestAction.CANCEL,
                                                request_id=str(request_id),
                                                requesting_device_id=self.client.device_id)
            await self.client.send_to_device(EventType.ROOM_KEY_REQUEST,
                                             {user_id: {device_id: cancel
                                                        for device_id in devices}
                                              for user_id, devices in from_devices.items()})
        return got_keys

    async def _receive_forwarded_room_key(self, evt: DecryptedOlmEvent) -> None:
        key: ForwardedRoomKeyEventContent = evt.content
        if await self.crypto_store.has_group_session(key.room_id, key.sender_key, key.session_id):
            self.log.debug(f"Ignoring received session {key.session_id} from {evt.sender}/"
                           f"{evt.sender_device}, as crypto store says we have it already")
            return
        key.forwarding_key_chain.append(evt.sender_key)
        sess = InboundGroupSession.import_session(key.session_key, key.signing_key, key.sender_key,
                                                  key.room_id, key.forwarding_key_chain)
        if key.session_id != sess.id:
            self.log.warning(f"Mismatched session ID while importing forwarded key from "
                             f"{evt.sender}/{evt.sender_device}: '{key.session_id}' != '{sess.id}'")
            return

        await self.crypto_store.put_group_session(key.room_id, key.sender_key, key.session_id, sess)
        self._mark_session_received(key.session_id)

        self.log.debug(f"Imported {key.session_id} for {key.room_id} "
                       f"from {evt.sender}/{evt.sender_device}")

        try:
            task = self._key_request_waiters[key.session_id]
        except KeyError:
            return
        task.set_result((evt.sender, evt.sender_device))



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Optional
from hashlib import sha256
import hmac
import json
import base64


def _get_checksum(key: str, payload: bytes) -> str:
    hasher = hmac.new(key.encode("utf-8"), msg=payload, digestmod=sha256)
    checksum = base64.urlsafe_b64encode(hasher.digest())
    return checksum.decode("utf-8").rstrip("=")


def sign_token(key: str, payload: Dict) -> str:
    payload_b64 = base64.urlsafe_b64encode(json.dumps(payload).encode("utf-8"))
    checksum = _get_checksum(key, payload_b64)
    payload_str = payload_b64.decode("utf-8").rstrip("=")
    return f"{checksum}:{payload_str}"


def verify_token(key: str, data: str) -> Optional[Dict]:
    if not data:
        return None

    try:
        checksum, payload = data.split(":", 1)
    except ValueError:
        return None

    payload += (3 - (len(payload) + 3) % 4) * "="

    if checksum != _get_checksum(key, payload.encode("utf-8")):
        return None

    payload = base64.urlsafe_b64decode(payload).decode("utf-8")
    try:
        return json.loads(payload)
    except json.JSONDecodeError:
        return None



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .handler import (command_handler, CommandEvent, SECTION_ADMIN)


@command_handler(needs_admin=True, needs_auth=False, help_section=SECTION_ADMIN,
                 help_text="Reset the bridge's megolm session in this room")
async def discard_megolm_session(evt: CommandEvent) -> None:
    if not evt.bridge.matrix.e2ee:
        await evt.reply("End-to-bridge encryption is not enabled on this bridge instance")
        return
    await evt.bridge.matrix.e2ee.crypto_store.remove_outbound_group_session(evt.room_id)
    await evt.reply("Successfully removed outbound group session for this room")



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, NamedTuple
from attr import dataclass

from .event import Membership
from .primitive import UserID, ContentURI
from .util import SerializableAttrs


@dataclass
class Member(SerializableAttrs['Member']):
    membership: Membership = None
    avatar_url: ContentURI = None
    displayname: str = None


@dataclass
class User(SerializableAttrs['User']):
    user_id: UserID
    avatar_url: ContentURI = None
    displayname: str = None


UserSearchResults = NamedTuple("UserSearchResults", results=List[User], limit=int)



from .account import OlmAccount
from .sessions import Session, InboundGroupSession, OutboundGroupSession
from .types import TrustState, DeviceIdentity, DecryptedOlmEvent
from .store import CryptoStore, StateStore, PgCryptoStore, PickleCryptoStore, MemoryCryptoStore
from .machine import OlmMachine
from .key_share import RejectKeyShare



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# Partly based on github.com/Cadair/python-appservice-framework (MIT license)
from typing import Optional, Callable, Awaitable, List, Set
from json import JSONDecodeError
from aiohttp import web
import asyncio
import logging

from mautrix.types import JSON, UserID, RoomAlias, Event, EphemeralEvent, SerializerError

QueryFunc = Callable[[web.Request], Awaitable[Optional[web.Response]]]
HandlerFunc = Callable[[Event], Awaitable]


class AppServiceServerMixin:
    loop: asyncio.AbstractEventLoop
    log: logging.Logger

    hs_token: str
    ephemeral_events: bool

    query_user: Callable[[UserID], JSON]
    query_alias: Callable[[RoomAlias], JSON]

    transactions: Set[str]
    event_handlers: List[HandlerFunc]

    def __init__(self, ephemeral_events: bool = False) -> None:
        self.transactions = set()
        self.event_handlers = []
        self.ephemeral_events = ephemeral_events

        async def default_query_handler(_):
            return None

        self.query_user = default_query_handler
        self.query_alias = default_query_handler

    def register_routes(self, app: web.Application) -> None:
        app.router.add_route("PUT", "/transactions/{transaction_id}",
                             self._http_handle_transaction)
        app.router.add_route("GET", "/rooms/{alias}", self._http_query_alias)
        app.router.add_route("GET", "/users/{user_id}", self._http_query_user)
        app.router.add_route("PUT", "/_matrix/app/v1/transactions/{transaction_id}",
                             self._http_handle_transaction)
        app.router.add_route("GET", "/_matrix/app/v1/rooms/{alias}", self._http_query_alias)
        app.router.add_route("GET", "/_matrix/app/v1/users/{user_id}", self._http_query_user)

    def _check_token(self, request: web.Request) -> bool:
        try:
            token = request.rel_url.query["access_token"]
        except KeyError:
            try:
                token = request.headers["Authorization"].lstrip("Bearer ")
            except (KeyError, AttributeError):
                return False

        if token != self.hs_token:
            return False

        return True

    async def _http_query_user(self, request: web.Request) -> web.Response:
        if not self._check_token(request):
            return web.json_response({"error": "Invalid auth token"}, status=401)

        try:
            user_id = request.match_info["user_id"]
        except KeyError:
            return web.json_response({"error": "Missing user_id parameter"}, status=400)

        try:
            response = await self.query_user(user_id)
        except Exception:
            self.log.exception("Exception in user query handler")
            return web.json_response({"error": "Internal appservice error"}, status=500)

        if not response:
            return web.json_response({}, status=404)
        return web.json_response(response)

    async def _http_query_alias(self, request: web.Request) -> web.Response:
        if not self._check_token(request):
            return web.json_response({"error": "Invalid auth token"}, status=401)

        try:
            alias = request.match_info["alias"]
        except KeyError:
            return web.json_response({"error": "Missing alias parameter"}, status=400)

        try:
            response = await self.query_alias(alias)
        except Exception:
            self.log.exception("Exception in alias query handler")
            return web.json_response({"error": "Internal appservice error"}, status=500)

        if not response:
            return web.json_response({}, status=404)
        return web.json_response(response)

    async def _http_handle_transaction(self, request: web.Request) -> web.Response:
        if not self._check_token(request):
            return web.json_response({"error": "Invalid auth token"}, status=401)

        transaction_id = request.match_info["transaction_id"]
        if transaction_id in self.transactions:
            return web.json_response({})

        try:
            json = await request.json()
        except JSONDecodeError:
            return web.json_response({"error": "Body is not JSON"}, status=400)

        try:
            events = json["events"]
        except KeyError:
            return web.json_response({"error": "Missing events object in body"}, status=400)

        if self.ephemeral_events:
            try:
                ephemeral = json["ephemeral"]
            except KeyError:
                try:
                    ephemeral = json["de.sorunome.msc2409.ephemeral"]
                except KeyError:
                    ephemeral = None
        else:
            ephemeral = None

        try:
            await self.handle_transaction(transaction_id, events=events, ephemeral=ephemeral)
        except Exception:
            self.log.exception("Exception in transaction handler")

        self.transactions.add(transaction_id)

        return web.json_response({})

    @staticmethod
    def _fix_prev_content(raw_event: JSON) -> None:
        try:
            raw_event["unsigned"]["prev_content"]
        except KeyError:
            try:
                raw_event.setdefault("unsigned", {})["prev_content"] = raw_event["prev_content"]
            except KeyError:
                pass

    async def handle_transaction(self, txn_id: str, events: List[JSON],
                                 ephemeral: Optional[List[JSON]] = None) -> None:
        for raw_edu in ephemeral or []:
            try:
                edu = EphemeralEvent.deserialize(raw_edu)
            except SerializerError:
                self.log.exception("Failed to deserialize ephemeral event %s", raw_edu)
            else:
                self.handle_matrix_event(edu)
        for raw_event in events:
            try:
                self._fix_prev_content(raw_event)
                event = Event.deserialize(raw_event)
            except SerializerError:
                self.log.exception("Failed to deserialize event %s", raw_event)
            else:
                self.handle_matrix_event(event)

    def handle_matrix_event(self, event: Event) -> None:
        if event.type.is_state and event.state_key is None:
            self.log.debug(f"Not sending {event.event_id} to handlers: expected state_key.")
            return

        async def try_handle(handler_func: HandlerFunc):
            try:
                await handler_func(event)
            except Exception:
                self.log.exception("Exception in Matrix event handler")

        for handler in self.event_handlers:
            asyncio.create_task(try_handle(handler))

    def matrix_event_handler(self, func: HandlerFunc) -> HandlerFunc:
        self.event_handlers.append(func)
        return func



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Tuple
import warnings
import asyncio
import json

from yarl import URL
from aiohttp import ClientSession, ContentTypeError, ClientError

from mautrix.api import HTTPAPI, Method
from mautrix.types import UserID, DeviceID, VersionsResponse, SerializerError
from mautrix.util.logging import TraceLogger
from mautrix.errors import (WellKnownNotURL, WellKnownNotJSON, WellKnownMissingHomeserver,
                            WellKnownUnexpectedStatus, WellKnownUnsupportedScheme,
                            WellKnownInvalidVersionsResponse)


class BaseClientAPI:
    """
    BaseClientAPI is the base class for :class:`ClientAPI`. This is separate from the main
    ClientAPI class so that the ClientAPI methods can be split into multiple classes (that
    inherit this class).All those section-specific method classes are inherited by the main
    ClientAPI class to create the full class.
    """

    localpart: str
    domain: str
    _mxid: UserID
    device_id: DeviceID
    api: HTTPAPI
    loop: asyncio.AbstractEventLoop
    log: TraceLogger

    def __init__(self, mxid: UserID = "", device_id: DeviceID = "", api: HTTPAPI = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None, *args, **kwargs) -> None:
        """
        Initialize a ClientAPI. You must either provide the

        Args:
            mxid: The Matrix ID of the user. This is used for things like setting profile metadata.
                Additionally, the homeserver domain is extracted from this string and used for
                setting aliases and such. This can be changed later using `set_mxid`.
            api: The :class:`HTTPAPI` instance to use. You can also pass the ``args`` and ``kwargs``
                to create a HTTPAPI instance rather than creating the instance yourself.``
        """
        if mxid:
            self.mxid = mxid
        else:
            self._mxid = None
            self.localpart = None
            self.domain = None
        self.device_id = device_id
        if loop:
            kwargs["loop"] = loop
        self.api = api or HTTPAPI(*args, **kwargs)
        self.loop = self.api.loop
        self.log = self.api.log

    @classmethod
    def parse_mxid(cls, mxid: UserID) -> Tuple[str, str]:
        warnings.warn("parse_mxid is deprecated, use parse_user_id instead",
                      category=DeprecationWarning)
        return cls.parse_user_id(mxid)

    @classmethod
    def parse_user_id(cls, mxid: UserID) -> Tuple[str, str]:
        """
        Parse the localpart and server name from a Matrix user ID.

        Args:
            mxid: The Matrix user ID.

        Returns:
            A tuple of (localpart, server_name).

        Raises:
            ValueError: if the given user ID is invalid.
        """
        if len(mxid) == 0:
            raise ValueError("User ID is empty")
        elif mxid[0] != "@":
            raise ValueError("User IDs start with @")
        try:
            sep = mxid.index(":")
        except ValueError as e:
            raise ValueError("User ID must contain domain separator") from e
        if sep == len(mxid) - 1:
            raise ValueError("User ID must contain domain")
        return mxid[1:sep], mxid[sep + 1:]

    @property
    def mxid(self) -> UserID:
        return self._mxid

    @mxid.setter
    def mxid(self, mxid: UserID) -> None:
        self.localpart, self.domain = self.parse_user_id(mxid)
        self._mxid = mxid

    async def versions(self) -> VersionsResponse:
        """Get client-server spec versions supported by the server."""
        resp = await self.api.request(Method.GET, "_matrix/client/versions")
        return VersionsResponse.deserialize(resp)

    @classmethod
    async def discover(cls, domain: str, session: Optional[ClientSession] = None) -> Optional[URL]:
        """
        Follow the server discovery spec to find the actual URL when given a Matrix server name.

        Args:
            domain: The server name (end of user ID) to discover.
            session: Optionally, the aiohttp ClientSession object to use.

        Returns:
            The parsed URL if the discovery succeeded.
            ``None`` if the request returned a 404 status.

        Raises:
            WellKnownError: for other errors
        """
        if session is None:
            async with ClientSession(headers={"User-Agent": HTTPAPI.default_ua}) as sess:
                return await cls._discover(domain, sess)
        else:
            return await cls._discover(domain, session)

    @classmethod
    async def _discover(cls, domain: str, session: ClientSession) -> Optional[URL]:
        well_known = URL.build(scheme="https", host=domain, path="/.well-known/matrix/client")
        async with session.get(well_known) as resp:
            if resp.status == 404:
                return None
            elif resp.status != 200:
                raise WellKnownUnexpectedStatus(resp.status)
            try:
                data = await resp.json(content_type=None)
            except (json.JSONDecodeError, ContentTypeError) as e:
                raise WellKnownNotJSON() from e

        try:
            homeserver_url = data["m.homeserver"]["base_url"]
        except KeyError as e:
            raise WellKnownMissingHomeserver() from e
        parsed_url = URL(homeserver_url)
        if not parsed_url.is_absolute():
            raise WellKnownNotURL()
        elif parsed_url.scheme not in ("http", "https"):
            raise WellKnownUnsupportedScheme(parsed_url.scheme)

        try:
            async with session.get(parsed_url / "_matrix/client/versions") as resp:
                data = VersionsResponse.deserialize(await resp.json())
                assert len(data.versions) > 0
        except (ClientError, json.JSONDecodeError, SerializerError, AssertionError) as e:
            raise WellKnownInvalidVersionsResponse() from e

        return parsed_url



# Copyright (c) 2021 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Any, Dict

from mautrix.types import JSON, Serializable, SerializableEnum, ExtensibleEnum


class RoomType(ExtensibleEnum):
    SPACE: 'RoomType'


class EventType(Serializable):
    class Class(SerializableEnum):
        UNKNOWN = "unknown"
        STATE = "state"
        MESSAGE = "message"
        ACCOUNT_DATA = "account_data"
        EPHEMERAL = "ephemeral"
        TO_DEVICE = "to_device"

    by_event_type: Dict[str, 'EventType']

    ROOM_CANONICAL_ALIAS: 'EventType'
    ROOM_ALIASES: 'EventType'
    ROOM_CREATE: 'EventType'
    ROOM_JOIN_RULES: 'EventType'
    ROOM_MEMBER: 'EventType'
    ROOM_POWER_LEVELS: 'EventType'
    ROOM_HISTORY_VISIBILITY: 'EventType'
    ROOM_NAME: 'EventType'
    ROOM_TOPIC: 'EventType'
    ROOM_AVATAR: 'EventType'
    ROOM_PINNED_EVENTS: 'EventType'
    ROOM_TOMBSTONE: 'EventType'
    ROOM_ENCRYPTION: 'EventType'

    SPACE_CHILD: 'EventType'
    SPACE_PARENT: 'EventType'

    ROOM_REDACTION: 'EventType'
    ROOM_MESSAGE: 'EventType'
    ROOM_ENCRYPTED: 'EventType'
    STICKER: 'EventType'
    REACTION: 'EventType'

    CALL_INVITE: 'EventType'
    CALL_CANDIDATES: 'EventType'
    CALL_SELECT_ANSWER: 'EventType'
    CALL_ANSWER: 'EventType'
    CALL_HANGUP: 'EventType'
    CALL_REJECT: 'EventType'
    CALL_NEGOTIATE: 'EventType'

    RECEIPT: 'EventType'
    TYPING: 'EventType'
    PRESENCE: 'EventType'

    DIRECT: 'EventType'
    PUSH_RULES: 'EventType'
    TAG: 'EventType'
    IGNORED_USER_LIST: 'EventType'

    TO_DEVICE_ENCRYPTED: 'EventType'
    ROOM_KEY: 'EventType'
    ROOM_KEY_WITHHELD: 'EventType'
    ORG_MATRIX_ROOM_KEY_WITHHELD: 'EventType'
    ROOM_KEY_REQUEST: 'EventType'
    FORWARDED_ROOM_KEY: 'EventType'

    ALL: 'EventType'

    is_message: bool
    is_state: bool
    is_ephemeral: bool
    is_account_data: bool
    is_to_device: bool

    t: str
    t_class: Class

    @classmethod
    def find(cls, t: str, t_class: Optional[Class] = None) -> 'EventType': ...

    def serialize(self) -> JSON: ...

    @classmethod
    def deserialize(cls, raw: JSON) -> Any: ...

    def with_class(self, t_class: Class) -> 'EventType': ...



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Dict, List, Optional
from attr import dataclass

from .primitive import UserID, DeviceID, IdentityKey, SigningKey
from .util import SerializableAttrs
from .event.encrypted import EncryptionAlgorithm, EncryptionKeyAlgorithm


@dataclass
class UnsignedDeviceInfo(SerializableAttrs['UnsignedDeviceInfo']):
    device_display_name: Optional[str] = None


@dataclass
class DeviceKeys(SerializableAttrs['DeviceKeys']):
    user_id: UserID
    device_id: DeviceID
    algorithms: List[EncryptionAlgorithm]
    keys: Dict[str, str]
    signatures: Dict[UserID, Dict[str, str]]
    unsigned: UnsignedDeviceInfo = None

    def __attrs_post_init__(self) -> None:
        if self.unsigned is None:
            self.unsigned = UnsignedDeviceInfo()

    @property
    def ed25519(self) -> SigningKey:
        try:
            return self.keys[f"{EncryptionKeyAlgorithm.ED25519}:{self.device_id}"]
        except KeyError:
            return None

    @property
    def curve25519(self) -> IdentityKey:
        try:
            return self.keys[f"{EncryptionKeyAlgorithm.CURVE25519}:{self.device_id}"]
        except KeyError:
            return None


@dataclass
class QueryKeysResponse(SerializableAttrs['QueryKeysResponse']):
    failures: Dict[str, Any]
    device_keys: Dict[UserID, Dict[DeviceID, DeviceKeys]]


@dataclass
class ClaimKeysResponse(SerializableAttrs['ClaimKeysResponse']):
    failures: Dict[str, Any]
    one_time_keys: Dict[UserID, Dict[DeviceID, Dict[str, Any]]]



__all__ = ["formatter", "logging", "config", "signed_token", "simple_template", "manhole",
           "markdown", "simple_lock", "file_store", "program", "async_db", "db", "opt_prometheus",
           "color_log", "magic"]



from .config import BaseBridgeConfig
from .custom_puppet import CustomPuppetMixin, CustomPuppetError, OnlyLoginSelf, InvalidAccessToken
from .matrix import BaseMatrixHandler
from .portal import BasePortal
from .user import BaseUser
from .puppet import BasePuppet
from .bridge import Bridge
from .notification_disabler import NotificationDisabler
from .async_getter_lock import async_getter_lock



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Callable, Awaitable, Optional
from abc import ABC

from mautrix.types import RoomID, UserID, RoomEncryptionStateEventContent
from mautrix.crypto import StateStore

from mautrix.bridge.portal import BasePortal

GetPortalFunc = Callable[[RoomID], Awaitable[BasePortal]]


class BaseCryptoStateStore(StateStore, ABC):
    get_portal: GetPortalFunc

    def __init__(self, get_portal: GetPortalFunc):
        self.get_portal = get_portal

    async def is_encrypted(self, room_id: RoomID) -> bool:
        portal = await self.get_portal(room_id)
        return portal.encrypted if portal else False


try:
    from mautrix.client.state_store.sqlalchemy import UserProfile, RoomState


    class SQLCryptoStateStore(BaseCryptoStateStore):
        @staticmethod
        async def find_shared_rooms(user_id: UserID) -> List[RoomID]:
            return [profile.user_id for profile in UserProfile.find_rooms_with_user(user_id)]

        @staticmethod
        async def get_encryption_info(room_id: RoomID) -> Optional[RoomEncryptionStateEventContent]:
            state = RoomState.get(room_id)
            if not state:
                return None
            return state.encryption
except ImportError:
    UserProfile = None
    RoomState = None
    SQLCryptoStateStore = None

try:
    from mautrix.util.async_db import Database


    class PgCryptoStateStore(BaseCryptoStateStore):
        db: Database

        def __init__(self, db: Database, get_portal: GetPortalFunc) -> None:
            super().__init__(get_portal)
            self.db = db

        async def find_shared_rooms(self, user_id: UserID) -> List[RoomID]:
            rows = await self.db.fetch("SELECT room_id FROM mx_user_profile "
                                       "LEFT JOIN portal ON portal.mxid=mx_user_profile.room_id "
                                       "WHERE user_id=$1 AND portal.encrypted=true", user_id)
            return [row["room_id"] for row in rows]

        async def get_encryption_info(self, room_id: RoomID
                                      ) -> Optional[RoomEncryptionStateEventContent]:
            val = await self.db.fetchval("SELECT encryption FROM mx_room_state WHERE room_id=$1",
                                         room_id)
            if not val:
                return None
            return RoomEncryptionStateEventContent.parse_json(val)
except ImportError:
    Database = None
    PgStateStore = None



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, NamedTuple, Optional, Union, TYPE_CHECKING

from mautrix.appservice import IntentAPI
from mautrix.errors import MatrixRequestError
from mautrix.types import RoomID, UserID, EventID, EventType

from .handler import command_handler, CommandEvent, SECTION_ADMIN
from ..portal import BasePortal

if TYPE_CHECKING:
    from .. import Bridge


class ManagementRoom(NamedTuple):
    room_id: RoomID
    user_id: UserID


class RoomSearchResults(NamedTuple):
    management_rooms: List[ManagementRoom]
    unidentified_rooms: List[RoomID]
    tombstoned_rooms: List[RoomID]
    portals: List[BasePortal]
    empty_portals: List[BasePortal]


async def _find_rooms(bridge: 'Bridge', intent: Optional[IntentAPI] = None) -> RoomSearchResults:
    results = RoomSearchResults([], [], [], [], [])
    intent = intent or bridge.az.intent
    rooms = await intent.get_joined_rooms()
    for room_id in rooms:
        portal = await bridge.get_portal(room_id)
        if not portal:
            try:
                tombstone = await intent.get_state_event(room_id, EventType.ROOM_TOMBSTONE)
                if tombstone and tombstone.replacement_room:
                    results.tombstoned_rooms.append(room_id)
                    continue
            except MatrixRequestError:
                pass
            try:
                members = await intent.get_room_members(room_id)
            except MatrixRequestError:
                members = []
            if len(members) == 2:
                other_member = members[0] if members[0] != intent.mxid else members[1]
                if bridge.is_bridge_ghost(other_member):
                    results.unidentified_rooms.append(room_id)
                else:
                    results.management_rooms.append(ManagementRoom(room_id, other_member))
            else:
                results.unidentified_rooms.append(room_id)
        else:
            members = await portal.get_authenticated_matrix_users()
            if len(members) == 0:
                results.empty_portals.append(portal)
            else:
                results.portals.append(portal)

    return results


@command_handler(needs_admin=True, needs_auth=False, management_only=True, name="clean-rooms",
                 help_section=SECTION_ADMIN,
                 help_text="Clean up unused portal/management rooms.")
async def clean_rooms(evt: CommandEvent) -> EventID:
    results = await _find_rooms(evt.bridge)

    reply = ["#### Management rooms (M)"]
    reply += ([f"{n + 1}. [M{n + 1}](https://matrix.to/#/{room}) (with {other_member}"
               for n, (room, other_member) in enumerate(results.management_rooms)]
              or ["No management rooms found."])
    reply.append("#### Active portal rooms (A)")
    reply += ([f"{n + 1}. [A{n + 1}](https://matrix.to/#/{portal.mxid}) "
               f"(to remote chat \"{portal.name}\")"
               for n, portal in enumerate(results.portals)]
              or ["No active portal rooms found."])
    reply.append("#### Unidentified rooms (U)")
    reply += ([f"{n + 1}. [U{n + 1}](https://matrix.to/#/{room})"
               for n, room in enumerate(results.unidentified_rooms)]
              or ["No unidentified rooms found."])
    reply.append("#### Tombstoned rooms (T)")
    reply += ([f"{n + 1}. [T{n + 1}](https://matrix.to/#/{room})"
               for n, room in enumerate(results.tombstoned_rooms)]
              or ["No tombstoned rooms found."])
    reply.append("#### Inactive portal rooms (I)")
    reply += ([f"{n}. [I{n}](https://matrix.to/#/{portal.mxid}) "
               f"(to remote chat \"{portal.name}\")"
               for n, portal in enumerate(results.empty_portals)]
              or ["No inactive portal rooms found."])

    reply += ["#### Usage",
              ("To clean the recommended set of rooms (unidentified & inactive portals), "
               "type `$cmdprefix+sp clean-recommended`"),
              "",
              ("To clean other groups of rooms, type `$cmdprefix+sp clean-groups <letters>` "
               "where `letters` are the first letters of the group names (M, A, U, I, T)"),
              "",
              ("To clean specific rooms, type `$cmdprefix+sp clean-range <range>` "
               "where `range` is the range (e.g. `5-21`) prefixed with the first letter of"
               "the group name. (e.g. `I2-6`)"),
              "",
              ("Please note that you will have to re-run `$cmdprefix+sp clean-rooms` "
               "between each use of the commands above.")]

    evt.sender.command_status = {
        "next": lambda clean_evt: set_rooms_to_clean(clean_evt, results),
        "action": "Room cleaning",
    }

    return await evt.reply("\n".join(reply))


async def set_rooms_to_clean(evt, results: RoomSearchResults) -> None:
    command = evt.args[0]
    rooms_to_clean: List[Union[BasePortal, RoomID]] = []
    if command == "clean-recommended":
        rooms_to_clean += results.empty_portals
        rooms_to_clean += results.unidentified_rooms
    elif command == "clean-groups":
        if len(evt.args) < 2:
            return await evt.reply("**Usage:** $cmdprefix+sp clean-groups [M][A][U][I]")
        groups_to_clean = evt.args[1].upper()
        if "M" in groups_to_clean:
            rooms_to_clean += [room_id for (room_id, user_id) in results.management_rooms]
        if "A" in groups_to_clean:
            rooms_to_clean += results.portals
        if "U" in groups_to_clean:
            rooms_to_clean += results.unidentified_rooms
        if "I" in groups_to_clean:
            rooms_to_clean += results.empty_portals
        if "T" in groups_to_clean:
            rooms_to_clean += results.tombstoned_rooms
    elif command == "clean-range":
        try:
            clean_range = evt.args[1]
            group, clean_range = clean_range[0], clean_range[1:]
            start, end = clean_range.split("-")
            start, end = int(start), int(end)
            if group == "M":
                group = [room_id for (room_id, user_id) in results.management_rooms]
            elif group == "A":
                group = results.portals
            elif group == "U":
                group = results.unidentified_rooms
            elif group == "I":
                group = results.empty_portals
            elif group == "T":
                group = results.tombstoned_rooms
            else:
                raise ValueError("Unknown group")
            rooms_to_clean = group[start - 1:end]
        except (KeyError, ValueError):
            return await evt.reply(
                "**Usage:** $cmdprefix+sp clean-range <_M|A|U|I_><range>")
    else:
        return await evt.reply(f"Unknown room cleaning action `{command}`. "
                               "Use `$cmdprefix+sp cancel` to cancel room "
                               "cleaning.")

    evt.sender.command_status = {
        "next": lambda confirm: execute_room_cleanup(confirm, rooms_to_clean),
        "action": "Room cleaning",
    }
    await evt.reply(f"To confirm cleaning up {len(rooms_to_clean)} rooms, type "
                    "`$cmdprefix+sp confirm-clean`.")


async def execute_room_cleanup(evt, rooms_to_clean: List[Union[BasePortal, RoomID]]) -> None:
    if len(evt.args) > 0 and evt.args[0] == "confirm-clean":
        await evt.reply(f"Cleaning {len(rooms_to_clean)} rooms. "
                        "This might take a while.")
        cleaned = 0
        for room in rooms_to_clean:
            if isinstance(room, BasePortal):
                await room.cleanup_and_delete()
                cleaned += 1
            else:
                await BasePortal.cleanup_room(evt.az.intent, room, "Room deleted")
                cleaned += 1
        evt.sender.command_status = None
        await evt.reply(f"{cleaned} rooms cleaned up successfully.")
    else:
        await evt.reply("Room cleaning cancelled.")



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Any, Optional, TYPE_CHECKING

from mautrix.client.state_store import FileStateStore

from .memory import ASStateStore

if TYPE_CHECKING:
    from mautrix.util.file_store import PathLike, Filer


class FileASStateStore(FileStateStore, ASStateStore):
    def __init__(self, path: 'PathLike', filer: Optional['Filer'] = None, binary: bool = True,
                 save_interval: float = 60.0) -> None:
        FileStateStore.__init__(self, path, filer, binary, save_interval)
        ASStateStore.__init__(self)

    def serialize(self) -> Dict[str, Any]:
        return {
            "registered": self._registered,
            **super().serialize(),
        }

    def deserialize(self, data: Dict[str, Any]) -> None:
        self._registered = data["registered"]
        super().deserialize(data)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Union, Optional, List

from attr import dataclass
import attr

from ..util import SerializableAttrs, Obj, deserializer, ExtensibleEnum
from ..primitive import JSON, UserID, RoomID, SessionID, IdentityKey, SigningKey, DeviceID
from .encrypted import EncryptionAlgorithm, EncryptedOlmEventContent
from .base import EventType, BaseEvent


class RoomKeyWithheldCode(ExtensibleEnum):
    BLACKLISTED: 'RoomKeyWithheldCode' = "m.blacklisted"
    UNVERIFIED: 'RoomKeyWithheldCode' = "m.unverified"
    UNAUTHORIZED: 'RoomKeyWithheldCode' = "m.unauthorized"
    UNAVAILABLE: 'RoomKeyWithheldCode' = "m.unavailable"
    NO_OLM_SESSION: 'RoomKeyWithheldCode' = "m.no_olm"


@dataclass
class RoomKeyWithheldEventContent(SerializableAttrs['RoomKeyWithheldEventContent']):
    algorithm: EncryptionAlgorithm
    sender_key: IdentityKey
    code: RoomKeyWithheldCode
    reason: Optional[str] = None
    room_id: Optional[RoomID] = None
    session_id: Optional[SessionID] = None


@dataclass
class RoomKeyEventContent(SerializableAttrs['RoomKeyEventContent']):
    algorithm: EncryptionAlgorithm
    room_id: RoomID
    session_id: SessionID
    session_key: str


class KeyRequestAction(ExtensibleEnum):
    REQUEST: 'KeyRequestAction' = "request"
    CANCEL: 'KeyRequestAction' = "request_cancellation"


@dataclass
class RequestedKeyInfo(SerializableAttrs['RequestedKeyInfo']):
    algorithm: EncryptionAlgorithm
    room_id: RoomID
    sender_key: IdentityKey
    session_id: SessionID


@dataclass
class RoomKeyRequestEventContent(SerializableAttrs['RoomKeyRequestEventContent']):
    action: KeyRequestAction
    requesting_device_id: DeviceID
    request_id: str
    body: Optional[RequestedKeyInfo] = None


@dataclass
class ForwardedRoomKeyEventContent(RoomKeyEventContent,
                                   SerializableAttrs['ForwardedRoomKeyEventContent']):
    sender_key: IdentityKey
    signing_key: SigningKey = attr.ib(metadata={"json": "sender_claimed_ed25519_key"})
    forwarding_key_chain: List[str] = attr.ib(metadata={"json": "forwarding_curve25519_key_chain"})


ToDeviceEventContent = Union[Obj, EncryptedOlmEventContent, RoomKeyWithheldEventContent,
                             RoomKeyEventContent, RoomKeyRequestEventContent,
                             ForwardedRoomKeyEventContent]
to_device_event_content_map = {
    EventType.TO_DEVICE_ENCRYPTED: EncryptedOlmEventContent,
    EventType.ROOM_KEY_WITHHELD: RoomKeyWithheldEventContent,
    EventType.ROOM_KEY_REQUEST: RoomKeyRequestEventContent,
    EventType.ROOM_KEY: RoomKeyEventContent,
    EventType.FORWARDED_ROOM_KEY: ForwardedRoomKeyEventContent,
}


# TODO remaining account data event types


@dataclass
class ToDeviceEvent(BaseEvent, SerializableAttrs['ToDeviceEvent']):
    sender: UserID
    content: ToDeviceEventContent

    @classmethod
    def deserialize(cls, data: JSON) -> 'ToDeviceEvent':
        try:
            evt_type = EventType.find(data.get("type"), t_class=EventType.Class.TO_DEVICE)
            data.setdefault("content", {})["__mautrix_event_type"] = evt_type
        except ValueError:
            return Obj(**data)
        return super().deserialize(data)

    @staticmethod
    @deserializer(ToDeviceEventContent)
    def deserialize_content(data: JSON) -> ToDeviceEventContent:
        evt_type = data.pop("__mautrix_event_type", None)
        content_type = to_device_event_content_map.get(evt_type, None)
        if not content_type:
            return Obj(**data)
        return content_type.deserialize(data)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Dict, Any, List, TYPE_CHECKING
from collections import defaultdict
from abc import ABC, abstractmethod
import asyncio
import logging

from mautrix.appservice import AppService, IntentAPI
from mautrix.types import (RoomID, EventID, MessageEventContent, EventType, EncryptionAlgorithm,
                           RoomEncryptionStateEventContent as Encryption, UserID)
from mautrix.errors import MatrixError, MatrixRequestError, MNotFound
from mautrix.util.logging import TraceLogger
from mautrix.util.simple_lock import SimpleLock

if TYPE_CHECKING:
    from .user import BaseUser
    from .matrix import BaseMatrixHandler
    from .bridge import Bridge


class BasePortal(ABC):
    log: TraceLogger = logging.getLogger("mau.portal")
    _async_get_locks: Dict[Any, asyncio.Lock] = defaultdict(lambda: asyncio.Lock())
    az: AppService
    matrix: 'BaseMatrixHandler'
    bridge: 'Bridge'
    loop: asyncio.AbstractEventLoop
    main_intent: IntentAPI
    mxid: Optional[RoomID]
    name: Optional[str]
    encrypted: bool
    is_direct: bool
    backfill_lock: SimpleLock

    @abstractmethod
    async def save(self) -> None:
        pass

    @abstractmethod
    async def handle_matrix_message(self, sender: 'BaseUser', message: MessageEventContent,
                                    event_id: EventID) -> None:
        pass

    async def check_dm_encryption(self) -> Optional[bool]:
        try:
            evt = await self.main_intent.get_state_event(self.mxid, EventType.ROOM_ENCRYPTION)
            self.log.debug("Found existing encryption event in direct portal: %s", evt)
            if evt and evt.algorithm == EncryptionAlgorithm.MEGOLM_V1:
                self.encrypted = True
        except MNotFound:
            pass
        if self.is_direct and self.matrix.e2ee and (self.bridge.config["bridge.encryption.default"]
                                                    or self.encrypted):
            return await self.enable_dm_encryption()
        return None

    async def enable_dm_encryption(self) -> bool:
        self.log.debug("Inviting bridge bot to room for end-to-bridge encryption")
        try:
            await self.main_intent.invite_user(self.mxid, self.az.bot_mxid)
            await self.az.intent.join_room_by_id(self.mxid)
            if not self.encrypted:
                await self.main_intent.send_state_event(self.mxid, EventType.ROOM_ENCRYPTION,
                                                        Encryption(EncryptionAlgorithm.MEGOLM_V1))
        except Exception:
            self.log.warning(f"Failed to enable end-to-bridge encryption", exc_info=True)
            return False

        self.encrypted = True
        return True

    async def _send_message(self, intent: IntentAPI, content: MessageEventContent,
                            event_type: EventType = EventType.ROOM_MESSAGE, **kwargs) -> EventID:
        if self.encrypted and self.matrix.e2ee:
            if intent.api.is_real_user:
                content[intent.api.real_user_content_key] = True
            event_type, content = await self.matrix.e2ee.encrypt(self.mxid, event_type, content)
        return await intent.send_message_event(self.mxid, event_type, content, **kwargs)

    @property
    @abstractmethod
    def bridge_info_state_key(self) -> str:
        pass

    @property
    @abstractmethod
    def bridge_info(self) -> Dict[str, Any]:
        pass

    # region Matrix room cleanup

    @abstractmethod
    async def delete(self) -> None:
        pass

    @classmethod
    async def cleanup_room(cls, intent: IntentAPI, room_id: RoomID, message: str = "Cleaning room",
                           puppets_only: bool = False) -> None:
        try:
            members = await intent.get_room_members(room_id)
        except MatrixError:
            members = []
        for user_id in members:
            puppet = await cls.bridge.get_puppet(user_id, create=False)
            if user_id != intent.mxid and (not puppets_only or puppet):
                try:
                    if puppet:
                        await puppet.intent.leave_room(room_id)
                    else:
                        await intent.kick_user(room_id, user_id, message)
                except MatrixError:
                    pass
        try:
            await intent.leave_room(room_id)
        except MatrixError:
            cls.log.warning(f"Failed to leave room {room_id} when cleaning up room", exc_info=True)

    async def cleanup_portal(self, message: str, puppets_only: bool = False) -> None:
        await self.cleanup_room(self.main_intent, self.mxid, message, puppets_only)
        await self.delete()

    async def unbridge(self) -> None:
        await self.cleanup_portal("Room unbridged", puppets_only=True)

    async def cleanup_and_delete(self) -> None:
        await self.cleanup_portal("Portal deleted")

    async def get_authenticated_matrix_users(self) -> List[UserID]:
        """
        Get the list of Matrix user IDs who can be bridged. This is used to determine if the portal
        is empty (and should be cleaned up) or not. Bridges should override this to check that the
        users are either logged in or the portal has a relaybot.
        """
        try:
            members = await self.main_intent.get_room_members(self.mxid)
        except MatrixRequestError:
            return []
        return [member for member in members
                if (not self.bridge.is_bridge_ghost(member)
                    and member != self.az.bot_mxid)]

    # endregion



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import Formatter, LogRecord
from copy import copy

PREFIX = "\033["
RESET = PREFIX + "0m"
MAU_COLOR = PREFIX + "32m"  # green
AIOHTTP_COLOR = PREFIX + "36m"  # cyan
MXID_COLOR = PREFIX + "33m"  # yellow

LEVEL_COLORS = {
    "DEBUG": "37m",  # white
    "INFO": "36m",  # cyan
    "WARNING": "33;1m",  # yellow
    "ERROR": "31;1m",  # red
    "CRITICAL": f"37;1m{PREFIX}41m",  # white on red bg
}

LEVELNAME_OVERRIDE = {
    name: f"{PREFIX}{color}{name}{RESET}"
    for name, color in LEVEL_COLORS.items()
}


class ColorFormatter(Formatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _color_name(self, module: str) -> str:
        as_api = "mau.as.api"
        if module.startswith(as_api):
            return f"{MAU_COLOR}{as_api}{RESET}.{MXID_COLOR}{module[len(as_api) + 1:]}{RESET}"
        elif module.startswith("mau."):
            try:
                next_dot = module.index(".", len("mau."))
                return (MAU_COLOR + module[:next_dot] + RESET
                        + "." + MXID_COLOR + module[next_dot + 1:] + RESET)
            except ValueError:
                return MAU_COLOR + module + RESET
        elif module.startswith("aiohttp"):
            return AIOHTTP_COLOR + module + RESET
        return module

    def format(self, record: LogRecord):
        colored_record: LogRecord = copy(record)
        colored_record.name = self._color_name(record.name)
        colored_record.levelname = LEVELNAME_OVERRIDE.get(record.levelname, record.levelname)
        return super().format(colored_record)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional
import logging
import asyncio


class SimpleLock:
    _event: asyncio.Event
    log: Optional[logging.Logger]
    message: Optional[str]

    def __init__(self, message: Optional[str] = None, log: Optional[logging.Logger] = None) -> None:
        self._event = asyncio.Event()
        self._event.set()
        self.log = log
        self.message = message

    def __enter__(self) -> None:
        self._event.clear()

    async def __aenter__(self) -> None:
        self.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._event.set()

    def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self.__exit__(exc_type, exc_val, exc_tb)

    @property
    def locked(self) -> bool:
        return not self._event.is_set()

    async def wait(self, task: Optional[str] = None) -> None:
        if not self._event.is_set():
            if self.log and self.message:
                self.log.debug(self.message, task)
            await self._event.wait()



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

from attr import dataclass
import attr

from ..primitive import RoomID, UserID, EventID
from ..util import Obj, SerializableAttrs
from .type import EventType


@dataclass
class BaseUnsigned:
    """Base unsigned information."""
    age: int = None


@dataclass
class BaseEvent:
    """Base event class. The only things an event **must** have are content and event type."""
    content: Obj
    type: EventType


@dataclass
class BaseRoomEvent(BaseEvent):
    """Base room event class. Room events must have a room ID, event ID, sender and timestamp in
    addition to the content and type in the base event."""
    room_id: RoomID
    event_id: EventID
    sender: UserID
    timestamp: int = attr.ib(metadata={"json": "origin_server_ts"})


@dataclass
class GenericEvent(BaseEvent, SerializableAttrs['GenericEvent']):
    """
    An event class that contains all possible top-level event keys and uses generic Obj's for object
    keys (content and unsigned)
    """
    content: Obj
    type: EventType
    room_id: Optional[RoomID] = None
    event_id: Optional[EventID] = None
    sender: Optional[UserID] = None
    timestamp: Optional[int] = None
    state_key: Optional[str] = None
    unsigned: Obj = None
    redacts: Optional[EventID] = None



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Iterable, Awaitable, Optional, Type, Union, Tuple, AsyncIterable, Any
from itertools import chain
from time import time
import argparse
import inspect
import logging
import logging.config
import asyncio
import signal
import copy
import sys

from .config import BaseFileConfig, BaseValidatableConfig, ConfigValueError, BaseMissingError

try:
    import uvloop
except ImportError:
    uvloop = None

try:
    import prometheus_client as prometheus
except ImportError:
    prometheus = None


NewTask = Union[Awaitable[Any], Iterable[Awaitable[Any]], AsyncIterable[Awaitable[Any]]]
TaskList = Iterable[Awaitable[Any]]


class Program:
    """
    A generic main class for programs that handles argument parsing, config loading, logger setup
    and general startup/shutdown lifecycle.
    """

    loop: asyncio.AbstractEventLoop
    log: logging.Logger
    parser: argparse.ArgumentParser
    args: argparse.Namespace

    config_class: Type[BaseFileConfig]
    config: BaseFileConfig

    startup_actions: TaskList
    shutdown_actions: TaskList

    module: str
    name: str
    version: str
    command: str
    description: str

    def __init__(self, module: Optional[str] = None, name: Optional[str] = None,
                 description: Optional[str] = None, command: Optional[str] = None,
                 version: Optional[str] = None, config_class: Optional[Type[BaseFileConfig]] = None
                 ) -> None:
        if module:
            self.module = module
        if name:
            self.name = name
        if description:
            self.description = description
        if command:
            self.command = command
        if version:
            self.version = version
        if config_class:
            self.config_class = config_class
        self.startup_actions = []
        self.shutdown_actions = []
        self._automatic_prometheus = True

    def run(self) -> None:
        """
        Prepare and run the program. This is the main entrypoint and the only function that should
        be called manually.
        """
        self._prepare()
        self._run()

    def _prepare(self) -> None:
        start_ts = time()

        self.preinit()

        self.log.info(f"Initializing {self.name} {self.version}")
        try:
            self.prepare()
        except Exception:
            self.log.critical("Unexpected error in initialization", exc_info=True)
            sys.exit(1)
        end_ts = time()
        self.log.info(f"Initialization complete in {round(end_ts - start_ts, 2)} seconds")

    def preinit(self) -> None:
        """
        First part of startup: parse command-line arguments, load and update config, prepare logger.
        Exceptions thrown here will crash the program immediately. Asyncio must not be used at this
        stage, as the loop is only initialized later.
        """
        self.prepare_arg_parser()
        self.args = self.parser.parse_args()

        self.prepare_config()
        self.prepare_log()
        self.check_config()

    def prepare_arg_parser(self) -> None:
        """Pre-init lifecycle method. Extend this if you want custom command-line arguments."""
        self.parser = argparse.ArgumentParser(description=self.description, prog=self.command)
        self.parser.add_argument("-c", "--config", type=str, default="config.yaml",
                                 metavar="<path>", help="the path to your config file")
        self.parser.add_argument("-b", "--base-config", type=str,
                                 default=f"pkg://{self.module}/example-config.yaml",
                                 metavar="<path>", help="the path to the example config "
                                                        "(for automatic config updates)")
        self.parser.add_argument("-n", "--no-update", action="store_true",
                                 help="Don't save updated config to disk")

    def prepare_config(self) -> None:
        """Pre-init lifecycle method. Extend this if you want to customize config loading."""
        self.config = self.config_class(self.args.config, self.args.base_config)
        self.load_and_update_config()

    def load_and_update_config(self) -> None:
        self.config.load()
        try:
            self.config.update(save=not self.args.no_update)
        except BaseMissingError:
            self.log.exception("Failed to read base config. Try specifying the correct path with "
                               "the --base-config flag.")
            sys.exit(12)

    def check_config(self) -> None:
        """Pre-init lifecycle method. Extend this if you want to customize config validation."""
        if not isinstance(self.config, BaseValidatableConfig):
            return
        try:
            self.config.check_default_values()
        except ConfigValueError as e:
            self.log.fatal(f"Configuration error: {e}")
            sys.exit(11)

    def prepare_log(self) -> None:
        """Pre-init lifecycle method. Extend this if you want to customize logging setup."""
        logging.config.dictConfig(copy.deepcopy(self.config["logging"]))
        self.log = logging.getLogger("mau.init")

    def prepare(self) -> None:
        """
        Lifecycle method where the primary program initialization happens.
        Use this to fill startup_actions with async startup tasks.
        """
        self.prepare_loop()

    def prepare_loop(self) -> None:
        """Init lifecycle method where the asyncio event loop is created."""
        if uvloop is not None:
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
            self.log.debug("Using uvloop for asyncio")

        self.loop = asyncio.get_event_loop()

    def start_prometheus(self) -> None:
        try:
            enabled = self.config["metrics.enabled"]
            listen_port = self.config["metrics.listen_port"]
        except KeyError:
            return
        if not enabled:
            return
        elif not prometheus:
            self.log.warning("Metrics are enabled in config, "
                             "but prometheus_client is not installed")
            return
        prometheus.start_http_server(listen_port)

    def _run(self) -> None:
        signal.signal(signal.SIGINT, signal.default_int_handler)
        signal.signal(signal.SIGTERM, signal.default_int_handler)

        try:
            self.log.debug("Running startup actions...")
            start_ts = time()
            self.loop.run_until_complete(self.start())
            end_ts = time()
            self.log.info(f"Startup actions complete in {round(end_ts - start_ts, 2)} seconds, "
                          "now running forever")
            self._stop_task = self.loop.create_future()
            self.loop.run_until_complete(self._stop_task)
            self.log.debug("manual_stop() called, stopping...")
        except KeyboardInterrupt:
            self.log.debug("Interrupt received, stopping...")
        except Exception:
            self.log.critical("Unexpected error in main event loop", exc_info=True)
            sys.exit(2)
        self.prepare_stop()
        self.loop.run_until_complete(self.stop())
        self.prepare_shutdown()
        self.log.info("Everything stopped, shutting down")
        sys.exit(0)

    async def start(self) -> None:
        """
        First lifecycle method called inside the asyncio event loop. Extend this if you want more
        control over startup than just filling startup_actions in the prepare step.
        """
        await asyncio.gather(*(self.startup_actions or []))
        if self._automatic_prometheus:
            self.start_prometheus()

    def prepare_stop(self) -> None:
        """
        Lifecycle method that is called before awaiting :meth:`stop`.
        Useful for filling shutdown_actions.
        """

    async def stop(self) -> None:
        """
        Lifecycle method used to stop things that need awaiting to stop. Extend this if you want
        more control over shutdown than just filling shutdown_actions in the prepare_stop method.
        """
        await asyncio.gather(*(self.shutdown_actions or []))

    def prepare_shutdown(self) -> None:
        """Lifecycle method that is called right before ``sys.exit(0)``."""

    def manual_stop(self) -> None:
        """Tell the event loop to cleanly stop and run the stop lifecycle steps."""
        self._stop_task.set_result(None)

    def add_startup_actions(self, *actions: NewTask) -> None:
        self.startup_actions = self._add_actions(self.startup_actions, actions)

    def add_shutdown_actions(self, *actions: NewTask) -> None:
        self.shutdown_actions = self._add_actions(self.shutdown_actions, actions)

    @staticmethod
    async def _unpack_async_iterator(iterable: AsyncIterable[Awaitable[Any]]) -> None:
        tasks = []
        async for task in iterable:
            if inspect.isawaitable(task):
                tasks.append(asyncio.create_task(task))
        await asyncio.gather(*tasks)

    def _add_actions(self, to: TaskList, add: Tuple[NewTask, ...]) -> TaskList:
        for item in add:
            if inspect.isasyncgen(item):
                to.append(self._unpack_async_iterator(item))
            elif inspect.isawaitable(item):
                if isinstance(to, list):
                    to.append(item)
                else:
                    to = chain(to, [item])
            elif isinstance(item, list):
                if isinstance(to, list):
                    to += item
                else:
                    to = chain(to, item)
            else:
                to = chain(to, item)
        return to



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Union, Optional

from mautrix.api import Method, Path
from mautrix.types import Obj, EventType, AccountDataEventContent, JSON, RoomID, Serializable
from mautrix.types.event.account_data import account_data_event_content_map

from ..base import BaseClientAPI


class AccountDataMethods(BaseClientAPI):
    """
    Methods in section 13.9 Client Config of the spec. These methods are used for storing user-local
    data on the homeserver to synchronize client configuration across sessions.

    See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#id125>`__"""

    async def get_account_data(self, type: Union[EventType, str], room_id: Optional[RoomID] = None
                               ) -> JSON:
        """
        Get a specific account data event from the homeserver.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#get-matrix-client-r0-user-userid-account-data-type>`__

        Args:
            type: The type of the account data event to get.
            room_id: Optionally, the room ID to get per-room account data.

        Returns:
            The data in the event.
        """
        if isinstance(type, EventType) and not type.is_account_data:
            raise ValueError("Event type is not an account data event type")
        base_path = Path.user[self.mxid]
        if room_id:
            base_path = base_path.rooms[room_id]
        return await self.api.request(Method.GET, base_path.account_data[type])

    async def set_account_data(self, type: Union[EventType, str], data: AccountDataEventContent,
                               room_id: Optional[RoomID] = None) -> None:
        """
        Store account data on the homeserver.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#put-matrix-client-r0-user-userid-account-data-type>`__

        Args:
            type: The type of the account data event to set.
            data: The content to store in that account data event.
            room_id: Optionally, the room ID to set per-room account data.
        """
        if isinstance(type, EventType) and not type.is_account_data:
            raise ValueError("Event type is not an account data event type")
        base_path = Path.user[self.mxid]
        if room_id:
            base_path = base_path.rooms[room_id]
        await self.api.request(Method.PUT, base_path.account_data[type],
                               data.serialize() if isinstance(data, Serializable) else data)



from .handler import (HelpSection, HelpCacheKey, command_handler, CommandHandler, CommandProcessor,
                      CommandHandlerFunc, CommandEvent, SECTION_GENERAL, SECTION_ADMIN)
from .meta import cancel, unknown_command, help_cmd
from . import admin, crypto, clean_rooms

__all__ = ["HelpSection", "HelpCacheKey", "command_handler", "CommandHandler", "CommandProcessor",
           "CommandHandlerFunc", "CommandEvent", "SECTION_GENERAL", "SECTION_ADMIN"]



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Union, Awaitable, Optional, List, Dict
from abc import ABC, abstractmethod

from mautrix.types import (StateEvent, EventType, RoomID, UserID, PowerLevelStateEventContent,
                           MemberStateEventContent, RoomEncryptionStateEventContent, Member,
                           Membership)


class StateStore(ABC):
    async def open(self) -> None:
        pass

    async def close(self) -> None:
        await self.flush()

    async def flush(self) -> None:
        pass

    @abstractmethod
    async def get_member(self, room_id: RoomID, user_id: UserID) -> Optional[Member]:
        pass

    @abstractmethod
    async def set_member(self, room_id: RoomID, user_id: UserID,
                         member: Union[Member, MemberStateEventContent]) -> None:
        pass

    @abstractmethod
    async def set_membership(self, room_id: RoomID, user_id: UserID, membership: Membership
                             ) -> None:
        pass

    @abstractmethod
    async def get_members(self, room_id: RoomID) -> Optional[List[UserID]]:
        pass

    async def get_members_filtered(self, room_id: RoomID, not_prefix: str, not_suffix: str,
                                   not_id: str) -> Optional[List[UserID]]:
        """
        A filtered version of get_members that only returns user IDs that aren't operated by a
        bridge. This should return the same as :meth:`get_members`, except users where the user ID
        is equal to not_id OR it starts with not_prefix AND ends with not_suffix.

        The default implementation simply calls :meth:`get_members`, but databases can implement
        this more efficiently.

        Args:
            room_id: The room ID to find.
            not_prefix: The user ID prefix to disallow.
            not_suffix: The user ID suffix to disallow.
            not_id: The user ID to disallow.
        """
        members = await self.get_members(room_id)
        if members is None:
            return None
        return [user_id for user_id in members
                if user_id != not_id and (not user_id.startswith(not_prefix)
                                          and not user_id.endswith(not_suffix))]

    @abstractmethod
    async def set_members(self, room_id: RoomID,
                          members: Dict[UserID, Union[Member, MemberStateEventContent]]) -> None:
        pass

    @abstractmethod
    async def has_full_member_list(self, room_id: RoomID) -> bool:
        pass

    @abstractmethod
    async def has_power_levels_cached(self, room_id: RoomID) -> bool:
        pass

    @abstractmethod
    async def get_power_levels(self, room_id: RoomID) -> Optional[PowerLevelStateEventContent]:
        pass

    @abstractmethod
    async def set_power_levels(self, room_id: RoomID, content: PowerLevelStateEventContent
                               ) -> None:
        pass

    @abstractmethod
    async def has_encryption_info_cached(self, room_id: RoomID) -> bool:
        pass

    @abstractmethod
    async def is_encrypted(self, room_id: RoomID) -> Optional[bool]:
        pass

    @abstractmethod
    async def get_encryption_info(self, room_id: RoomID
                                  ) -> Optional[RoomEncryptionStateEventContent]:
        pass

    @abstractmethod
    async def set_encryption_info(self, room_id: RoomID, content: RoomEncryptionStateEventContent
                                  ) -> None:
        pass

    async def update_state(self, evt: StateEvent) -> None:
        if evt.type == EventType.ROOM_POWER_LEVELS:
            await self.set_power_levels(evt.room_id, evt.content)
        elif evt.type == EventType.ROOM_MEMBER:
            await self.set_member(evt.room_id, UserID(evt.state_key), evt.content)
        elif evt.type == EventType.ROOM_ENCRYPTION:
            await self.set_encryption_info(evt.room_id, evt.content)

    async def get_membership(self, room_id: RoomID, user_id: UserID) -> Membership:
        member = await self.get_member(room_id, user_id)
        return member.membership if member else Membership.LEAVE

    async def is_joined(self, room_id: RoomID, user_id: UserID) -> bool:
        return (await self.get_membership(room_id, user_id)) == Membership.JOIN

    def joined(self, room_id: RoomID, user_id: UserID) -> Awaitable[None]:
        return self.set_membership(room_id, user_id, Membership.JOIN)

    def invited(self, room_id: RoomID, user_id: UserID) -> Awaitable[None]:
        return self.set_membership(room_id, user_id, Membership.INVITE)

    def left(self, room_id: RoomID, user_id: UserID) -> Awaitable[None]:
        return self.set_membership(room_id, user_id, Membership.LEAVE)

    async def has_power_level(self, room_id: RoomID, user_id: UserID, event_type: EventType
                              ) -> Optional[bool]:
        room_levels = await self.get_power_levels(room_id)
        if not room_levels:
            return None
        return room_levels.get_user_level(user_id) >= room_levels.get_event_level(event_type)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Callable
from abc import ABC

from ruamel.yaml.comments import CommentedMap

from .base import BaseConfig
from .recursive_dict import RecursiveDict


class BaseProxyConfig(BaseConfig, ABC):
    def __init__(self, load: Callable[[], CommentedMap],
                 load_base: Callable[[], Optional[RecursiveDict[CommentedMap]]],
                 save: Callable[[RecursiveDict[CommentedMap]], None]) -> None:
        super().__init__()
        self._data = CommentedMap()
        self._load_proxy = load
        self._load_base_proxy = load_base
        self._save_proxy = save

    def load(self) -> None:
        self._data = self._load_proxy() or CommentedMap()

    def load_base(self) -> Optional[RecursiveDict[CommentedMap]]:
        return self._load_base_proxy()

    def save(self) -> None:
        self._save_proxy(self._data)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Union, Optional, TYPE_CHECKING

from mautrix.types import (RoomID, UserID, Member, Membership, PowerLevelStateEventContent,
                           RoomEncryptionStateEventContent, MemberStateEventContent)
from mautrix.util.file_store import FileStore

from .memory import MemoryStateStore

if TYPE_CHECKING:
    from mautrix.util.file_store import PathLike, Filer


class FileStateStore(FileStore, MemoryStateStore):
    def __init__(self, path: 'PathLike', filer: Optional['Filer'] = None, binary: bool = True,
                 save_interval: float = 60.0) -> None:
        FileStore.__init__(self, path, filer, binary, save_interval)
        MemoryStateStore.__init__(self)

    async def set_membership(self, room_id: RoomID, user_id: UserID,
                             membership: Membership) -> None:
        await super().set_membership(room_id, user_id, membership)
        self._time_limited_flush()

    async def set_member(self, room_id: RoomID, user_id: UserID,
                         member: Union[Member, MemberStateEventContent]) -> None:
        await super().set_member(room_id, user_id, member)
        self._time_limited_flush()

    async def set_encryption_info(self, room_id: RoomID,
                                  content: RoomEncryptionStateEventContent) -> None:
        await super().set_encryption_info(room_id, content)
        self._time_limited_flush()

    async def set_power_levels(self, room_id: RoomID, content: PowerLevelStateEventContent) -> None:
        await super().set_power_levels(room_id, content)
        self._time_limited_flush()



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Iterable, Dict

from sqlalchemy import Column, Text, Enum

from mautrix.types import RoomID, UserID, ContentURI, Member, Membership
from mautrix.util.db import Base

from .mx_room_state import RoomState


class UserProfile(Base):
    __tablename__ = "mx_user_profile"

    room_id: RoomID = Column(Text, primary_key=True)
    user_id: UserID = Column(Text, primary_key=True)
    membership: Membership = Column(Enum(Membership), nullable=False, default=Membership.LEAVE)
    displayname: str = Column(Text, nullable=True)
    avatar_url: ContentURI = Column(Text, nullable=True)

    def member(self) -> Member:
        return Member(membership=self.membership, displayname=self.displayname,
                      avatar_url=self.avatar_url)

    @classmethod
    def get(cls, room_id: RoomID, user_id: UserID) -> Optional['UserProfile']:
        return cls._select_one_or_none((cls.c.room_id == room_id) & (cls.c.user_id == user_id))

    @classmethod
    def all_in_room(cls, room_id: RoomID, prefix: str = None, suffix: str = None, bot: str = None
                    ) -> Iterable['UserProfile']:
        clause = (((cls.c.membership == Membership.JOIN)
                   | (cls.c.membership == Membership.INVITE))
                  & (cls.c.room_id == room_id))
        if bot:
            clause = clause & (cls.c.user_id != bot)
        if prefix:
            clause = clause & ~cls.c.user_id.startswith(prefix, autoescape=True)
        if suffix:
            clause = clause & ~cls.c.user_id.startswith(suffix, autoescape=True)
        return cls._select_all(clause)

    @classmethod
    def find_rooms_with_user(cls, user_id: UserID) -> Iterable['UserProfile']:
        return cls._select_all((cls.c.user_id == user_id)
                               & (cls.c.room_id == RoomState.c.room_id)
                               & (RoomState.c.is_encrypted == True))

    @classmethod
    def delete_all(cls, room_id: RoomID) -> None:
        with cls.db.begin() as conn:
            conn.execute(cls.t.delete().where(cls.c.room_id == room_id))

    @classmethod
    def bulk_replace(cls, room_id: RoomID, members: Dict[UserID, Member]) -> None:
        with cls.db.begin() as conn:
            delete_condition = cls.c.room_id == room_id
            cls.db.execute(cls.t.delete().where(delete_condition))
            conn.execute(cls.t.insert(),
                         [dict(room_id=room_id, user_id=user_id, membership=member.membership,
                               displayname=member.displayname, avatar_url=member.avatar_url)
                          for user_id, member in members.items()])



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Dict, Optional

from mautrix.types import UserID, SyncToken, DeviceID, DeviceKeys
from mautrix.errors import DeviceValidationError

from .base import BaseOlmMachine, verify_signature_json
from .types import DeviceIdentity, TrustState


class DeviceListMachine(BaseOlmMachine):
    async def _fetch_keys(self, users: List[UserID], since: SyncToken = "",
                          include_untracked: bool = False
                          ) -> Dict[UserID, Dict[DeviceID, DeviceIdentity]]:
        if not include_untracked:
            users = await self.crypto_store.filter_tracked_users(users)
        if len(users) == 0:
            return {}
        users = set(users)

        self.log.trace(f"Querying keys for {users}")
        keys = await self.client.query_keys(users, token=since)

        for server, err in keys.failures.items():
            self.log.warning(f"Query keys failure for {server}: {err}")

        data = {}
        for user_id, devices in keys.device_keys.items():
            users.remove(user_id)

            new_devices = {}
            existing_devices = (await self.crypto_store.get_devices(user_id)) or {}

            self.log.trace(f"Updating devices for {user_id}, got {len(devices)}, "
                           f"have {len(existing_devices)} in store")
            changed = False
            for device_id, keys in devices.items():
                try:
                    existing = existing_devices[device_id]
                except KeyError:
                    existing = None
                    changed = True
                self.log.trace(f"Validating device {keys} of {user_id}")
                try:
                    new_device = await self._validate_device(user_id, device_id, keys, existing)
                except DeviceValidationError as e:
                    self.log.warning(f"Failed to validate device {device_id} of {user_id}: {e}")
                else:
                    if new_device:
                        new_devices[device_id] = new_device
            self.log.debug(f"Storing new device list for {user_id} "
                           f"containing {len(new_devices)} devices")
            await self.crypto_store.put_devices(user_id, new_devices)
            data[user_id] = new_devices

            if changed or len(new_devices) != len(existing_devices):
                await self.on_devices_changed(user_id)

        for user_id in users:
            self.log.warning(f"Didn't get any keys for user {user_id}")

        return data

    async def get_or_fetch_device(self, user_id: UserID, device_id: DeviceID
                                  ) -> Optional[DeviceIdentity]:
        device = await self.crypto_store.get_device(user_id, device_id)
        if device is not None:
            return device
        devices = await self._fetch_keys([user_id], include_untracked=True)
        try:
            return devices[user_id][device_id]
        except KeyError:
            return None

    async def on_devices_changed(self, user_id: UserID) -> None:
        shared_rooms = await self.state_store.find_shared_rooms(user_id)
        self.log.debug(f"Devices of {user_id} changed, "
                       f"invalidating group session in {shared_rooms}")
        await self.crypto_store.remove_outbound_group_sessions(shared_rooms)

    @staticmethod
    async def _validate_device(user_id: UserID, device_id: DeviceID, device_keys: DeviceKeys,
                               existing: Optional[DeviceIdentity] = None) -> DeviceIdentity:
        if user_id != device_keys.user_id:
            raise DeviceValidationError("mismatching user ID in parameter and keys object")
        elif device_id != device_keys.device_id:
            raise DeviceValidationError("mismatching device ID in parameter and keys object")

        signing_key = device_keys.ed25519
        if not signing_key:
            raise DeviceValidationError("didn't find ed25519 signing key")
        identity_key = device_keys.curve25519
        if not identity_key:
            raise DeviceValidationError("didn't find curve25519 identity key")

        if existing and existing.signing_key != signing_key:
            raise DeviceValidationError("received update for device with different signing key")

        if not verify_signature_json(device_keys.serialize(), user_id, device_id, signing_key):
            raise DeviceValidationError("invalid signature on device keys")

        name = device_keys.unsigned.device_display_name or device_id

        return DeviceIdentity(user_id=user_id, device_id=device_id, identity_key=identity_key,
                              signing_key=signing_key, trust=TrustState.UNSET, name=name,
                              deleted=False)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional
from abc import ABC, abstractmethod

from ruamel.yaml.comments import CommentedMap

from .recursive_dict import RecursiveDict


class BaseMissingError(ValueError):
    pass


class ConfigUpdateHelper:
    base: RecursiveDict[CommentedMap]

    def __init__(self, base: RecursiveDict, config: RecursiveDict) -> None:
        self.base = base
        self.source = config

    def copy(self, from_path: str, to_path: Optional[str] = None) -> None:
        if from_path in self.source:
            self.base[to_path or from_path] = self.source[from_path]

    def copy_dict(self, from_path: str, to_path: Optional[str] = None,
                  override_existing_map: Optional[bool] = True) -> None:
        if from_path in self.source:
            to_path = to_path or from_path
            if override_existing_map or to_path not in self.base:
                self.base[to_path] = CommentedMap()
            for key, value in self.source[from_path].items():
                self.base[to_path][key] = value

    def __iter__(self):
        yield self.copy
        yield self.copy_dict
        yield self.base


class BaseConfig(ABC, RecursiveDict[CommentedMap]):
    @abstractmethod
    def load(self) -> None:
        pass

    @abstractmethod
    def load_base(self) -> Optional[RecursiveDict[CommentedMap]]:
        pass

    def load_and_update(self) -> None:
        self.load()
        self.update()

    @abstractmethod
    def save(self) -> None:
        pass

    def update(self, save: bool = True) -> None:
        base = self.load_base()
        if not base:
            raise BaseMissingError("Can't update() without base config")

        self.do_update(ConfigUpdateHelper(base, self))
        self._data = base._data
        if save:
            self.save()

    @abstractmethod
    def do_update(self, helper: ConfigUpdateHelper) -> None:
        pass



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

from mautrix.types import EventID, UserID, ContentURI
from mautrix.errors import MatrixRequestError, IntentError, MForbidden

from .handler import (command_handler, CommandEvent, SECTION_ADMIN)


@command_handler(needs_admin=True, needs_auth=False, name="set-pl",
                 help_section=SECTION_ADMIN, help_args="<_level_> [_mxid_]",
                 help_text="Set a temporary power level without affecting the remote platform.")
async def set_power_level(evt: CommandEvent) -> EventID:
    try:
        level = int(evt.args[0])
    except (KeyError, IndexError):
        return await evt.reply("**Usage:** `$cmdprefix+sp set-pl <level> [mxid]`")
    except ValueError:
        return await evt.reply("The level must be an integer.")
    mxid = evt.args[1] if len(evt.args) > 1 else evt.sender.mxid
    levels = await evt.main_intent.get_power_levels(evt.room_id, ignore_cache=True)
    levels.users[mxid] = level
    try:
        return await evt.main_intent.set_power_levels(evt.room_id, levels)
    except MForbidden as e:
        await evt.reply(f"I don't seem to have permission to update power levels: {e.message}")
    except (MatrixRequestError, IntentError):
        evt.log.exception("Failed to update power levels")
        return await evt.reply("Failed to update power levels (see logs for more details)")


@command_handler(needs_admin=True, needs_auth=False, name="set-avatar",
                 help_section=SECTION_ADMIN, help_args="<_mxc:// uri_> [_mxid_]",
                 help_text="Set an avatar for a ghost user.")
async def set_ghost_avatar(evt: CommandEvent) -> Optional[EventID]:
    try:
        mxc_uri = ContentURI(evt.args[0])
    except (KeyError, IndexError):
        return await evt.reply("**Usage:** `$cmdprefix+sp set-avatar <mxc_uri> [mxid]`")
    if not mxc_uri.startswith("mxc://"):
        return await evt.reply("The URI has to start with mxc://.")
    if len(evt.args) > 1:
        # TODO support parsing mention pills instead of requiring a plaintext mxid
        puppet = await evt.processor.bridge.get_puppet(UserID(evt.args[1]))
        if puppet is None:
            return await evt.reply("The given mxid was not a valid ghost user.")
        intent = puppet.intent
    elif evt.is_portal:
        intent = evt.portal.main_intent
        if intent == evt.az.intent:
            return await evt.reply("No mxid given and the main intent is not a ghost user.")
    else:
        return await evt.reply("No mxid given and not in a portal.")
    try:
        return await intent.set_avatar_url(mxc_uri)
    except (MatrixRequestError, IntentError):
        evt.log.exception("Failed to set avatar.")
        return await evt.reply("Failed to set avatar (see logs for more details).")


@command_handler(needs_admin=True, needs_auth=False, name="remove-avatar",
                 help_section=SECTION_ADMIN, help_args="[_mxid_]",
                 help_text="Remove the avatar for a ghost user.")
async def remove_ghost_avatar(evt: CommandEvent) -> Optional[EventID]:
    if len(evt.args) > 0:
        puppet = await evt.processor.bridge.get_puppet(UserID(evt.args[0]))
        if puppet is None:
            return await evt.reply("The given mxid was not a valid ghost user.")
        intent = puppet.intent
    elif evt.is_portal:
        intent = evt.portal.main_intent
        if intent == evt.az.intent:
            return await evt.reply("No mxid given and the main intent is not a ghost user.")
    else:
        return await evt.reply("No mxid given and not in a portal.")
    try:
        return await intent.set_avatar_url(ContentURI(""))
    except (MatrixRequestError, IntentError):
        evt.log.exception("Failed to remove avatar.")
        return await evt.reply("Failed to remove avatar (see logs for more details).")


@command_handler(needs_admin=True, needs_auth=False, name="set-displayname",
                 help_section=SECTION_ADMIN, help_args="<_displayname_> [_mxid_]",
                 help_text="Set the display name for a ghost user.")
async def set_ghost_display_name(evt: CommandEvent) -> Optional[EventID]:
    if len(evt.args) > 1:
        # This allows whitespaces in the name
        puppet = await evt.processor.bridge.get_puppet(UserID(evt.args[len(evt.args) - 1]))
        if puppet is None:
            return await evt.reply("The given mxid was not a valid ghost user. "
                                   "If the display name has whitespaces mxid is required")
        intent = puppet.intent
        displayname = " ".join(evt.args[:-1])
    elif evt.is_portal:
        intent = evt.portal.main_intent
        if intent == evt.az.intent:
            return await evt.reply("No mxid given and the main intent is not a ghost user.")
        displayname = evt.args[0]
    else:
        return await evt.reply("No mxid given and not in a portal.")
    try:
        return await intent.set_displayname(displayname)
    except (MatrixRequestError, IntentError):
        evt.log.exception("Failed to set display name.")
        return await evt.reply("Failed to set display name (see logs for more details).")


@command_handler(needs_admin=True, needs_auth=False, name="remove-displayname",
                 help_section=SECTION_ADMIN, help_args="[_mxid_]",
                 help_text="Remove the display name for a ghost user.")
async def set_ghost_display_name(evt: CommandEvent) -> Optional[EventID]:
    if len(evt.args) > 0:
        puppet = await evt.processor.bridge.get_puppet(UserID(evt.args[0]))
        if puppet is None:
            return await evt.reply("The given mxid was not a valid ghost user.")
        intent = puppet.intent
    elif evt.is_portal:
        intent = evt.portal.main_intent
        if intent == evt.az.intent:
            return await evt.reply("No mxid given and the main intent is not a ghost user.")
    else:
        return await evt.reply("No mxid given and not in a portal (see logs for more details).")
    try:
        return await intent.set_displayname("")
    except (MatrixRequestError, IntentError):
        evt.log.exception("Failed to remove display name.")
        return await evt.reply("Failed to remove display name (see logs for more details).")



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Dict, List, Awaitable, Iterator
from abc import ABC, abstractmethod
from itertools import chain
import asyncio
import logging
import hashlib
import hmac
import json

from yarl import URL
from aiohttp import ClientConnectionError

from mautrix.types import (UserID, FilterID, Filter, RoomEventFilter, RoomFilter, EventFilter,
                           EventType, SyncToken, RoomID, Event, PresenceState, StateFilter,
                           LoginType)
from mautrix.appservice import AppService, IntentAPI
from mautrix.errors import (IntentError, MatrixError, MatrixRequestError, MatrixInvalidToken,
                            WellKnownError)
from mautrix.api import Path

from .matrix import BaseMatrixHandler


class CustomPuppetError(MatrixError):
    """Base class for double puppeting setup errors."""


class InvalidAccessToken(CustomPuppetError):
    def __init__(self):
        super().__init__("The given access token was invalid.")


class OnlyLoginSelf(CustomPuppetError):
    def __init__(self):
        super().__init__("You may only replace your puppet with your own Matrix account.")


class HomeserverURLNotFound(CustomPuppetError):
    def __init__(self, domain: str):
        super().__init__(f"Could not discover a valid homeserver URL for {domain}")


class OnlyLoginTrustedDomain(CustomPuppetError):
    def __init__(self):
        super().__init__("This bridge doesn't allow double-puppeting "
                         "with accounts on untrusted servers.")


class CustomPuppetMixin(ABC):
    """
    Mixin for the Puppet class to enable Matrix puppeting.

    Attributes:
        sync_with_custom_puppets: Whether or not custom puppets should /sync
        allow_discover_url: Allow logging into other homeservers using .well-known discovery.
        homeserver_url_map: Static map from server name to URL that are always allowed to log in.
        only_handle_own_synced_events: Whether or not typing notifications and read receipts by
                                       other users should be filtered away before passing them to
                                       the Matrix event handler.

        az: The AppService object.
        loop: The asyncio event loop.
        log: The logger to use.
        mx: The Matrix event handler to send /sync events to.

        by_custom_mxid: A mapping from custom mxid to puppet object.

        default_mxid: The default user ID of the puppet.
        default_mxid_intent: The IntentAPI for the default user ID.
        custom_mxid: The user ID of the custom puppet.
        access_token: The access token for the custom puppet.

        intent: The primary IntentAPI.
    """

    sync_with_custom_puppets: bool = True
    allow_discover_url: bool = False
    homeserver_url_map: Dict[str, URL] = {}
    only_handle_own_synced_events: bool = True
    login_shared_secret_map: Dict[str, bytes] = {}
    login_device_name: Optional[str] = None

    az: AppService
    loop: asyncio.AbstractEventLoop
    log: logging.Logger
    mx: BaseMatrixHandler

    by_custom_mxid: Dict[UserID, 'CustomPuppetMixin'] = {}

    default_mxid: UserID
    default_mxid_intent: IntentAPI
    custom_mxid: Optional[UserID]
    access_token: Optional[str]
    base_url: Optional[URL]
    next_batch: Optional[SyncToken]

    intent: IntentAPI

    _sync_task: Optional[asyncio.Future] = None

    @abstractmethod
    async def save(self) -> None:
        """Save the information of this puppet. Called from :meth:`switch_mxid`"""

    @property
    def mxid(self) -> UserID:
        """The main Matrix user ID of this puppet."""
        return self.custom_mxid or self.default_mxid

    @property
    def is_real_user(self) -> bool:
        """Whether or not this puppet uses a real Matrix user instead of an appservice-owned ID."""
        return bool(self.custom_mxid and self.access_token)

    def _fresh_intent(self) -> IntentAPI:
        return (self.az.intent.user(self.custom_mxid, self.access_token, self.base_url)
                if self.is_real_user else self.default_mxid_intent)

    @classmethod
    def can_auto_login(cls, mxid: UserID) -> bool:
        _, server = cls.az.intent.parse_user_id(mxid)
        return server in cls.login_shared_secret_map and (server in cls.homeserver_url_map
                                                          or server == cls.az.domain)

    @classmethod
    async def _login_with_shared_secret(cls, mxid: UserID) -> Optional[str]:
        _, server = cls.az.intent.parse_user_id(mxid)
        try:
            secret = cls.login_shared_secret_map[server]
        except KeyError:
            return None
        try:
            base_url = cls.homeserver_url_map[server]
        except KeyError:
            if server == cls.az.domain:
                base_url = cls.az.intent.api.base_url
            else:
                return None
        password = hmac.new(secret, mxid.encode("utf-8"), hashlib.sha512).hexdigest()
        url = base_url / str(Path.login)
        resp = await cls.az.http_session.post(url, data=json.dumps({
            "type": str(LoginType.PASSWORD),
            "initial_device_display_name": cls.login_device_name,
            "device_id": cls.login_device_name,
            "identifier": {
                "type": "m.id.user",
                "user": mxid,
            },
            "password": password
        }), headers={"Content-Type": "application/json"})
        data = await resp.json()
        return data["access_token"]

    async def switch_mxid(self, access_token: Optional[str], mxid: Optional[UserID],
                          start_sync_task: bool = True) -> None:
        """
        Switch to a real Matrix user or away from one.

        Args:
            access_token: The access token for the custom account, or ``None`` to switch back to
                          the appservice-owned ID.
            mxid: The expected Matrix user ID of the custom account, or ``None`` when
                  ``access_token`` is None.
            start_sync_task: Whether or not syncing should be started after logging in.
        """
        if access_token == "auto":
            access_token = await self._login_with_shared_secret(mxid)
            if not access_token:
                raise ValueError("Failed to log in with shared secret")
            self.log.debug(f"Logged in for {mxid} using shared secret")

        if mxid is not None:
            _, mxid_domain = self.az.intent.parse_user_id(mxid)
            if mxid_domain in self.homeserver_url_map:
                base_url = self.homeserver_url_map[mxid_domain]
            elif mxid_domain == self.az.domain:
                base_url = None
            else:
                if not self.allow_discover_url:
                    raise OnlyLoginTrustedDomain()
                try:
                    base_url = await IntentAPI.discover(mxid_domain, self.az.http_session)
                except WellKnownError as e:
                    raise HomeserverURLNotFound(mxid_domain) from e
                if base_url is None:
                    raise HomeserverURLNotFound(mxid_domain)
        else:
            base_url = None

        prev_mxid = self.custom_mxid
        self.custom_mxid = mxid
        self.access_token = access_token
        self.base_url = base_url
        self.intent = self._fresh_intent()

        await self.start(start_sync_task=start_sync_task)

        try:
            del self.by_custom_mxid[prev_mxid]
        except KeyError:
            pass
        if self.mxid != self.default_mxid:
            self.by_custom_mxid[self.mxid] = self
            try:
                await self._leave_rooms_with_default_user()
            except Exception:
                self.log.warning("Error when leaving rooms with default user", exc_info=True)
        await self.save()

    async def try_start(self, retry_auto_login: bool = True) -> None:
        try:
            await self.start(retry_auto_login=retry_auto_login)
        except Exception:
            self.log.exception("Failed to initialize custom mxid")

    async def start(self, retry_auto_login: bool = False, start_sync_task: bool = True) -> None:
        """Initialize the custom account this puppet uses. Should be called at startup to start
        the /sync task. Is called by :meth:`switch_mxid` automatically."""
        if not self.is_real_user:
            return

        try:
            mxid = await self.intent.whoami()
        except MatrixInvalidToken as e:
            if retry_auto_login and self.custom_mxid and self.can_auto_login(self.custom_mxid):
                self.log.debug(f"Got {e.errcode} while trying to initialize custom mxid")
                await self.switch_mxid("auto", self.custom_mxid, start_sync_task=start_sync_task)
                return
            self.log.warning(f"Got {e.errcode} while trying to initialize custom mxid")
            mxid = None
        if not mxid or mxid != self.custom_mxid:
            if self.custom_mxid and self.by_custom_mxid.get(self.custom_mxid) == self:
                del self.by_custom_mxid[self.custom_mxid]
            self.custom_mxid = None
            self.access_token = None
            self.next_batch = None
            await self.save()
            self.intent = self._fresh_intent()
            if mxid != self.custom_mxid:
                raise OnlyLoginSelf()
            raise InvalidAccessToken()
        if self.sync_with_custom_puppets and start_sync_task:
            if self._sync_task:
                self._sync_task.cancel()
            self.log.info(f"Initialized custom mxid: {mxid}. Starting sync task")
            self._sync_task = asyncio.create_task(self._try_sync())
        else:
            self.log.info(f"Initialized custom mxid: {mxid}. Not starting sync task")

    def stop(self) -> None:
        """Cancel the sync task."""
        if self._sync_task:
            self._sync_task.cancel()
            self._sync_task = None

    async def default_puppet_should_leave_room(self, room_id: RoomID) -> bool:
        """
        Whether or not the default puppet user should leave the given room when this puppet is
        switched to using a custom user account.

        Args:
            room_id: The room to check.

        Returns:
            Whether or not the default user account should leave.
        """
        return True

    async def _leave_rooms_with_default_user(self) -> None:
        for room_id in await self.default_mxid_intent.get_joined_rooms():
            try:
                if await self.default_puppet_should_leave_room(room_id):
                    await self.default_mxid_intent.leave_room(room_id)
                    await self.intent.ensure_joined(room_id)
            except (IntentError, MatrixRequestError):
                pass

    def _create_sync_filter(self) -> Awaitable[FilterID]:
        all_events = EventType.find("*")
        return self.intent.create_filter(Filter(
            account_data=EventFilter(types=[all_events]),
            presence=EventFilter(
                types=[EventType.PRESENCE],
                senders=[self.custom_mxid] if self.only_handle_own_synced_events else None,
            ),
            room=RoomFilter(
                include_leave=False,
                state=StateFilter(not_types=[all_events]),
                timeline=RoomEventFilter(not_types=[all_events]),
                account_data=RoomEventFilter(not_types=[all_events]),
                ephemeral=RoomEventFilter(types=[
                    EventType.TYPING,
                    EventType.RECEIPT,
                ]),
            ),
        ))

    def _filter_events(self, room_id: RoomID, events: List[Dict]) -> Iterator[Event]:
        for event in events:
            event["room_id"] = room_id
            if self.only_handle_own_synced_events:
                # We only want events about the custom puppet user, but we can't use
                # filters for typing and read receipt events.
                evt_type = EventType.find(event.get("type", None))
                event.setdefault("content", {})
                if evt_type == EventType.TYPING:
                    is_typing = self.custom_mxid in event["content"].get("user_ids", [])
                    event["content"]["user_ids"] = [self.custom_mxid] if is_typing else []
                elif evt_type == EventType.RECEIPT:
                    try:
                        event_id, receipt = event["content"].popitem()
                        data = receipt["m.read"][self.custom_mxid]
                        event["content"] = {event_id: {"m.read": {self.custom_mxid: data}}}
                    except KeyError:
                        continue
            yield event

    def _handle_sync(self, sync_resp: Dict) -> None:
        # Get events from rooms -> join -> [room_id] -> ephemeral -> events (array)
        ephemeral_events = (
            event
            for room_id, data in sync_resp.get("rooms", {}).get("join", {}).items()
            for event in self._filter_events(room_id, data.get("ephemeral", {}).get("events", []))
        )

        # Get events from presence -> events (array)
        presence_events = sync_resp.get("presence", {}).get("events", [])

        # Deserialize and handle all events
        for event in chain(ephemeral_events, presence_events):
            asyncio.create_task(self.mx.try_handle_sync_event(Event.deserialize(event)))

    async def _try_sync(self) -> None:
        try:
            await self._sync()
        except asyncio.CancelledError:
            self.log.info(f"Syncing for {self.custom_mxid} cancelled")
        except Exception:
            self.log.critical(f"Fatal error syncing {self.custom_mxid}", exc_info=True)

    async def _sync(self) -> None:
        if not self.is_real_user:
            self.log.warning("Called sync() for non-custom puppet.")
            return
        custom_mxid: UserID = self.custom_mxid
        access_token_at_start: str = self.access_token
        errors: int = 0
        filter_id: FilterID = await self._create_sync_filter()
        self.log.debug(f"Starting syncer for {custom_mxid} with sync filter {filter_id}.")
        while access_token_at_start == self.access_token:
            try:
                cur_batch = self.next_batch
                sync_resp = await self.intent.sync(filter_id=filter_id, since=cur_batch,
                                                   set_presence=PresenceState.OFFLINE)
                try:
                    self.next_batch = sync_resp.get("next_batch", None)
                except Exception:
                    self.log.warning("Failed to store next batch", exc_info=True)
                errors = 0
                if cur_batch is not None:
                    self._handle_sync(sync_resp)
            except MatrixInvalidToken:
                # TODO when not using syncing, we should still check this occasionally and relogin
                self.log.warning(f"Access token for {custom_mxid} got invalidated, restarting...")
                await self.start(retry_auto_login=True, start_sync_task=False)
                if self.is_real_user:
                    self.log.info("Successfully relogined custom puppet, continuing sync")
                    filter_id = await self._create_sync_filter()
                    access_token_at_start = self.access_token
                else:
                    self.log.warning("Something went wrong during relogin")
                    raise
            except (MatrixError, ClientConnectionError, asyncio.TimeoutError) as e:
                errors += 1
                wait = min(errors, 11) ** 2
                self.log.warning(f"Syncer for {custom_mxid} errored: {e}. "
                                 f"Waiting for {wait} seconds...")
                await asyncio.sleep(wait)
        self.log.debug(f"Syncer for custom puppet {custom_mxid} stopped.")



# Copyright (c) 2021 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, List

from mautrix.api import Method, Path
from mautrix.types import (PushRule, PushRuleScope, PushRuleKind, PushRuleID, PushAction,
                           PushCondition)

from ..base import BaseClientAPI


class PushRuleMethods(BaseClientAPI):
    """
    Methods in section 13.13 Push Notifications of the spec. These methods are used for modifying
    what triggers push notifications.

    See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#id89>`__"""

    async def get_push_rule(self, scope: PushRuleScope, kind: PushRuleKind, rule_id: PushRuleID
                            ) -> PushRule:
        """
        Retrieve a single specified push rule.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#get-matrix-client-r0-pushrules-scope-kind-ruleid>`__

        Args:
            scope: The scope of the push rule.
            kind: The kind of rule.
            rule_id: The identifier of the rule.

        Returns:
            The push rule information.
        """
        resp = await self.api.request(Method.GET, Path.pushrules[scope][kind][rule_id])
        return PushRule.deserialize(resp)

    async def set_push_rule(self, scope: PushRuleScope, kind: PushRuleKind, rule_id: PushRuleID,
                            actions: List[PushAction], pattern: Optional[str] = None,
                            before: Optional[PushRuleID] = None, after: Optional[PushRuleID] = None,
                            conditions: List[PushCondition] = None) -> None:
        """
        Create or modify a push rule.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#put-matrix-client-r0-pushrules-scope-kind-ruleid>`__

        Args:
            scope: The scope of the push rule.
            kind: The kind of rule.
            rule_id: The identifier for the rule.
            before:
            after:
            actions: The actions to perform when the conditions for the rule are met.
            pattern: The glob-style pattern to match against for ``content`` rules.
            conditions: The conditions for the rule for ``underride`` and ``override`` rules.
        """
        query = {}
        if after:
            query["after"] = after
        if before:
            query["before"] = before
        content = {"actions": [act.serialize() for act in actions]}
        if conditions:
            content["conditions"] = [cond.serialize() for cond in conditions]
        if pattern:
            content["pattern"] = pattern
        await self.api.request(Method.PUT, Path.pushrules[scope][kind][rule_id],
                               query_params=query, content=content)

    async def remove_push_rule(self, scope: PushRuleScope, kind: PushRuleKind, rule_id: PushRuleID
                               ) -> None:
        """
        Remove a push rule.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1#delete-matrix-client-r0-pushrules-scope-kind-ruleid>`__

        Args:
            scope: The scope of the push rule.
            kind: The kind of rule.
            rule_id: The identifier of the rule.
        """
        await self.api.request(Method.DELETE, Path.pushrules[scope][kind][rule_id])



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import json

import olm

from mautrix.types import EncryptedMegolmEventContent, EncryptionAlgorithm, Event, EncryptedEvent
from mautrix.errors import (DecryptionError, SessionNotFound, DuplicateMessageIndex,
                            VerificationError, DecryptedPayloadError, MismatchingRoomError)

from .types import TrustState
from .base import BaseOlmMachine


class MegolmDecryptionMachine(BaseOlmMachine):
    async def decrypt_megolm_event(self, evt: EncryptedEvent) -> Event:
        """
        Decrypt an event that was encrypted using Megolm.

        Args:
            evt: The whole encrypted event.

        Returns:
            The decrypted event, including some unencrypted metadata from the input event.

        Raises:
            DecryptionError: If decryption failed.
        """
        if not isinstance(evt.content, EncryptedMegolmEventContent):
            raise DecryptionError("Unsupported event content class")
        elif evt.content.algorithm != EncryptionAlgorithm.MEGOLM_V1:
            raise DecryptionError("Unsupported event encryption algorithm")
        session = await self.crypto_store.get_group_session(evt.room_id, evt.content.sender_key,
                                                            evt.content.session_id)
        if session is None:
            # TODO check if olm session is wedged
            raise SessionNotFound(evt.content.session_id, evt.content.sender_key)
        try:
            plaintext, index = session.decrypt(evt.content.ciphertext)
        except olm.OlmGroupSessionError as e:
            raise DecryptionError("Failed to decrypt megolm event") from e
        if not await self.crypto_store.validate_message_index(evt.content.sender_key,
                                                              evt.content.session_id,
                                                              evt.event_id, index, evt.timestamp):
            raise DuplicateMessageIndex()

        verified = False
        if ((evt.content.device_id == self.client.device_id
             and session.signing_key == self.account.signing_key
             and evt.content.sender_key == self.account.identity_key)):
            verified = True
        else:
            device = await self.crypto_store.get_device(evt.sender, evt.content.device_id)
            if device and device.trust == TrustState.VERIFIED and not session.forwarding_chain:
                if ((device.signing_key != session.signing_key
                     or device.identity_key != evt.content.sender_key)):
                    raise VerificationError()
                verified = True
            # else: TODO query device keys?

        try:
            data = json.loads(plaintext)
            room_id = data["room_id"]
            event_type = data["type"]
            content = data["content"]
        except json.JSONDecodeError as e:
            raise DecryptedPayloadError("Failed to parse megolm payload") from e
        except KeyError as e:
            raise DecryptedPayloadError("Megolm payload is missing fields") from e

        if room_id != evt.room_id:
            raise MismatchingRoomError()

        result = Event.deserialize({
            "room_id": evt.room_id,
            "event_id": evt.event_id,
            "sender": evt.sender,
            "origin_server_ts": evt.timestamp,
            "type": event_type,
            "content": content,
        })
        result.unsigned = evt.unsigned
        if evt.content.relates_to:
            if hasattr(result.content, "relates_to"):
                if not result.content.relates_to:
                    result.content.relates_to = evt.content.relates_to
            elif "m.relates_to" not in result.content:
                result.content["m.relates_to"] = evt.content.relates_to.serialize()
        result.type = result.type.with_class(evt.type.t_class)
        result["mautrix"] = {
            "verified": verified,
        }
        return result



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Sequence, Union
from abc import ABC, abstractmethod
from enum import Enum, auto


class EntityType(Enum):
    """EntityType is a Matrix formatting entity type."""
    BOLD = auto()
    ITALIC = auto()
    STRIKETHROUGH = auto()
    UNDERLINE = auto()
    URL = auto()
    EMAIL = auto()
    USER_MENTION = auto()
    ROOM_MENTION = auto()
    PREFORMATTED = auto()
    INLINE_CODE = auto()
    BLOCKQUOTE = auto()
    HEADER = auto()


class FormattedString(ABC):
    """FormattedString is an abstract HTML parsing target."""

    @abstractmethod
    def append(self, *args: Union[str, 'FormattedString']) -> 'FormattedString':
        """
        Append strings to this FormattedString.

        This method may mutate the source object, but it is not required to do so.
        Make sure to always use the return value when mutating and to duplicate strings if you don't
        want the original to change.

        Args:
            *args: The strings to append.

        Returns:
            A FormattedString that is a concatenation of this string and the given strings.
        """
        pass

    @abstractmethod
    def prepend(self, *args: Union[str, 'FormattedString']) -> 'FormattedString':
        """
        Prepend strings to this FormattedString.

        This method may mutate the source object, but it is not required to do so.
        Make sure to always use the return value when mutating and to duplicate strings if you don't
        want the original to change.

        Args:
            *args: The strings to prepend.

        Returns:
            A FormattedString that is a concatenation of the given strings and this string.
        """
        pass

    @abstractmethod
    def format(self, entity_type: EntityType, **kwargs) -> 'FormattedString':
        """
        Apply formatting to this FormattedString.

        This method may mutate the source object, but it is not required to do so.
        Make sure to always use the return value when mutating and to duplicate strings if you don't
        want the original to change.

        Args:
            entity_type: The type of formatting to apply to this string.
            **kwargs: Additional metadata required by the formatting type.

        Returns:
            A FormattedString with the given formatting applied.
        """
        pass

    @abstractmethod
    def trim(self) -> 'FormattedString':
        """
        Trim surrounding whitespace from this FormattedString.

        This method may mutate the source object, but it is not required to do so.
        Make sure to always use the return value when mutating and to duplicate strings if you don't
        want the original to change.

        Returns:
            A FormattedString without surrounding whitespace.
        """
        pass

    @abstractmethod
    def split(self, separator, max_items: int = -1) -> List['FormattedString']:
        """
        Split this FormattedString by the given separator.

        Args:
            separator: The separator to split by.
            max_items: The maximum number of items to return. If the limit is reached, the remaining
                       string will be returned as one even if it contains the separator.

        Returns:
            The split strings.
        """
        pass

    @classmethod
    def concat(cls, *args: Union[str, 'FormattedString']) -> 'FormattedString':
        """
        Concatenate many FormattedStrings.

        Args:
            *args: The strings to concatenate.

        Returns:
            A FormattedString that is a concatenation of the given strings.
        """
        return cls.join(items=args, separator="")

    @classmethod
    @abstractmethod
    def join(cls, items: Sequence[Union[str, 'FormattedString']],
             separator: str = " ") -> 'FormattedString':
        """
        Join a list of FormattedStrings with the given separator.

        Args:
            items: The strings to join.
            separator: The separator to join them with.

        Returns:
            A FormattedString that is a combination of the given strings with the given separator
            between each one.
        """
        pass



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Callable, Awaitable, Any, Union


class _NoopPrometheusEntity:
    """NoopPrometheusEntity is a class that can be used as a no-op placeholder for prometheus
    metrics objects when prometheus_client isn't installed."""

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        if not kwargs and len(args) == 1 and callable(args[0]):
            return args[0]
        return self

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __getattr__(self, item):
        return self


try:
    from prometheus_client import Counter, Gauge, Summary, Histogram, Info, Enum

    is_installed = True
except ImportError:
    Counter = Gauge = Summary = Histogram = Info = Enum = _NoopPrometheusEntity
    is_installed = False


def async_time(metric):
    if not hasattr(metric, "time") or not callable(metric.time):
        raise ValueError("async_time only supports metrics that support timing")

    def decorator(fn):
        async def wrapper(*args, **kwargs):
            with metric.time():
                return await fn(*args, **kwargs)

        return wrapper if is_installed else fn

    return decorator


__all__ = ["Counter", "Gauge", "Summary", "Histogram", "Info", "Enum", "async_time", "is_installed"]



# Copyright (c) 2021 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Union
import functools

import magic

try:
    _from_buffer = functools.partial(magic.from_buffer, mime=True)
    _from_filename = functools.partial(magic.from_file, mime=True)
except AttributeError:
    def _from_buffer(data: bytes) -> str:
        return magic.detect_from_content(data).mime_type


    def _from_filename(file: str) -> str:
        return magic.detect_from_filename(file).mime_type


def mimetype(data: Union[bytes, str]) -> str:
    if isinstance(data, str):
        return _from_filename(data)
    return _from_buffer(data)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import commonmark


class HtmlEscapingRenderer(commonmark.HtmlRenderer):
    def __init__(self, allow_html: bool = False):
        super().__init__()
        self.allow_html = allow_html

    def lit(self, s):
        if self.allow_html:
            return super().lit(s)
        return super().lit(s.replace("<", "&lt;").replace(">", "&gt;"))

    def image(self, node, entering):
        prev = self.allow_html
        self.allow_html = True
        super().image(node, entering)
        self.allow_html = prev


md_parser = commonmark.Parser()
yes_html_renderer = commonmark.HtmlRenderer()
no_html_renderer = HtmlEscapingRenderer()


def render(message: str, allow_html: bool = False) -> str:
    parsed = md_parser.parse(message)
    if allow_html:
        return yes_html_renderer.render(parsed)
    else:
        return no_html_renderer.render(parsed)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Dict, NewType, Union

from attr import dataclass

from ..util import SerializableAttrs, SerializableEnum, deserializer
from ..primitive import UserID, RoomID, EventID, JSON
from .type import EventType
from .base import BaseEvent, GenericEvent


@dataclass
class TypingEventContent(SerializableAttrs['TypingEventContent']):
    user_ids: List[UserID]


@dataclass
class TypingEvent(BaseEvent, SerializableAttrs['TypingEvent']):
    room_id: RoomID
    content: TypingEventContent


class PresenceState(SerializableEnum):
    ONLINE = "online"
    OFFLINE = "offline"
    UNAVAILABLE = "unavailable"


@dataclass
class PresenceEventContent(SerializableAttrs['PresenceEventContent']):
    presence: PresenceState
    last_active_ago: int = None
    status_msg: str = None
    currently_active: bool = None


@dataclass
class PresenceEvent(BaseEvent, SerializableAttrs['PresenceEvent']):
    sender: UserID
    content: PresenceEventContent


@dataclass
class SingleReceiptEventContent(SerializableAttrs['SingleReceiptEventContent']):
    ts: int


class ReceiptType(SerializableEnum):
    READ = "m.read"


ReceiptEventContent = Dict[EventID, Dict[ReceiptType, Dict[UserID, SingleReceiptEventContent]]]


@dataclass
class ReceiptEvent(BaseEvent, SerializableAttrs['ReceiptEvent']):
    room_id: RoomID
    content: ReceiptEventContent


EphemeralEvent = NewType("EphemeralEvent", Union[PresenceEvent, TypingEvent, ReceiptEvent])


@deserializer(EphemeralEvent)
def deserialize_ephemeral_event(data: JSON) -> EphemeralEvent:
    event_type = EventType.find(data.get("type", None))
    if event_type == EventType.RECEIPT:
        return ReceiptEvent.deserialize(data)
    elif event_type == EventType.TYPING:
        return TypingEvent.deserialize(data)
    elif event_type == EventType.PRESENCE:
        return PresenceEvent.deserialize(data)
    else:
        return GenericEvent.deserialize(data)


setattr(EphemeralEvent, "deserialize", deserialize_ephemeral_event)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Any, Optional, List, TYPE_CHECKING
from collections import defaultdict
from abc import ABC, abstractmethod
import logging
import asyncio

from mautrix.api import Method, UnstableClientPath
from mautrix.appservice import AppService
from mautrix.types import UserID, RoomID, EventType, Membership
from mautrix.errors import MNotFound
from mautrix.util.logging import TraceLogger
from mautrix.util.opt_prometheus import Gauge

from .portal import BasePortal

if TYPE_CHECKING:
    from .bridge import Bridge

AsmuxPath = UnstableClientPath["com.beeper.asmux"]


class BaseUser(ABC):
    log: TraceLogger = logging.getLogger("mau.user")
    _async_get_locks: Dict[Any, asyncio.Lock] = defaultdict(lambda: asyncio.Lock())
    az: AppService
    bridge: 'Bridge'
    loop: asyncio.AbstractEventLoop

    is_whitelisted: bool
    is_admin: bool
    mxid: UserID
    command_status: Optional[Dict[str, Any]]
    _metric_value: Dict[Gauge, bool]

    dm_update_lock: asyncio.Lock

    @abstractmethod
    async def is_logged_in(self) -> bool:
        raise NotImplementedError()

    async def is_in_portal(self, portal: BasePortal) -> bool:
        try:
            member_event = await portal.main_intent.get_state_event(
                portal.mxid, EventType.ROOM_MEMBER, self.mxid)
        except MNotFound:
            return False
        return member_event and member_event.membership in (Membership.JOIN, Membership.INVITE)

    async def get_direct_chats(self) -> Dict[UserID, List[RoomID]]:
        raise NotImplementedError()

    async def update_direct_chats(self, dms: Optional[Dict[UserID, List[RoomID]]] = None) -> None:
        """
        Update the m.direct account data of the user.

        Args:
            dms: DMs to _add_ to the list. If not provided, the list is _replaced_ with the result
                of :meth:`get_direct_chats`.
        """
        if not self.bridge.config["bridge.sync_direct_chat_list"]:
            return

        puppet = await self.bridge.get_double_puppet(self.mxid)
        if not puppet or not puppet.is_real_user:
            return

        self.log.debug("Updating m.direct list on homeserver")
        replace = dms is None
        dms = dms or await self.get_direct_chats()
        if self.bridge.config.get("homeserver.asmux", False):
            # This uses a secret endpoint for atomically updating the DM list
            await puppet.intent.api.request(Method.PUT if replace else Method.PATCH, AsmuxPath.dms,
                                            content=dms, headers={"X-Asmux-Auth": self.az.as_token})
        else:
            async with self.dm_update_lock:
                try:
                    current_dms = await puppet.intent.get_account_data(EventType.DIRECT)
                except MNotFound:
                    current_dms = {}
                if replace:
                    # Filter away all existing DM statuses with bridge users
                    current_dms = {user: rooms for user, rooms in current_dms.items()
                                   if not self.bridge.is_bridge_ghost(user)}
                # Add DM statuses for all rooms in our database
                current_dms.update(dms)
                await puppet.intent.set_account_data(EventType.DIRECT, current_dms)

    def _track_metric(self, metric: Gauge, value: bool) -> None:
        if self._metric_value[metric] != value:
            if value:
                metric.inc(1)
            else:
                metric.dec(1)
            self._metric_value[metric] = value



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Callable, Awaitable, Dict, TYPE_CHECKING
import functools
import asyncio
import json

import olm

from mautrix.types import (UserID, DeviceID, SigningKey, EncryptionKeyAlgorithm, SessionID,
                           RequestedKeyInfo, RoomID, IdentityKey)
from mautrix.client import Client
from mautrix.util.logging import TraceLogger

from .types import DeviceIdentity

if TYPE_CHECKING:
    from .store import CryptoStore, StateStore
    from .account import OlmAccount

    from typing import TypedDict


    class SignedObject(TypedDict):
        signatures: Dict[UserID, Dict[str, str]]
        unsigned: Any


class BaseOlmMachine:
    client: Client
    log: TraceLogger
    loop: asyncio.AbstractEventLoop
    crypto_store: 'CryptoStore'
    state_store: 'StateStore'

    account: 'OlmAccount'

    allow_unverified_devices: bool
    share_to_unverified_devices: bool
    allow_key_share: Callable[[DeviceIdentity, RequestedKeyInfo], Awaitable[bool]]

    # Futures that wait for responses to a key request
    _key_request_waiters: Dict[SessionID, asyncio.Future]
    # Futures that wait for a session to be received (either normally or through a key request)
    _inbound_session_waiters: Dict[SessionID, asyncio.Future]

    async def wait_for_session(self, room_id: RoomID, sender_key: IdentityKey,
                               session_id: SessionID, timeout: float = 3) -> bool:
        try:
            fut = self._inbound_session_waiters[session_id]
        except KeyError:
            fut = self._inbound_session_waiters[session_id] = self.loop.create_future()
        try:
            return await asyncio.wait_for(asyncio.shield(fut), timeout)
        except asyncio.TimeoutError:
            return await self.crypto_store.has_group_session(room_id, sender_key, session_id)

    def _mark_session_received(self, session_id: SessionID) -> None:
        try:
            self._inbound_session_waiters.pop(session_id).set_result(True)
        except KeyError:
            return


canonical_json = functools.partial(json.dumps, ensure_ascii=False, separators=(",", ":"),
                                   sort_keys=True)


def verify_signature_json(data: 'SignedObject', user_id: UserID, device_id: DeviceID,
                          key: SigningKey) -> bool:
    data_copy = {**data}
    data_copy.pop("unsigned", None)
    signatures = data_copy.pop("signatures")
    signature = signatures[user_id][f"{EncryptionKeyAlgorithm.ED25519}:{device_id}"]
    signed_data = canonical_json(data_copy)
    try:
        olm.ed25519_verify(key, signed_data, signature)
        return True
    except olm.OlmVerifyError:
        return False



from .api import ClientAPI
from .state_store import StateStore, MemoryStateStore, FileStateStore, SyncStore, MemorySyncStore
from .syncer import EventHandler, InternalEventType, SyncStream
from .dispatcher import MembershipEventDispatcher, Dispatcher, SimpleDispatcher
from .store_updater import StoreUpdatingAPI
from .encryption_manager import EncryptingAPI, DecryptionDispatcher
from .client import Client



from .base import MatrixError, MatrixConnectionError, MatrixResponseError, IntentError
from .request import (MatrixRequestError, MatrixUnknownRequestError, MatrixStandardRequestError,
                      standard_error, make_request_error,
                      MForbidden, MatrixInvalidToken, MUnknownToken, MMissingToken,
                      MatrixBadRequest, MatrixBadContent, MBadJSON, MNotJSON, MNotFound,
                      MLimitExceeded, MUnknown, MUnrecognized, MUnauthorized, MUserInUse,
                      MInvalidUsername, MRoomInUse, MInvalidRoomState, MUnsupportedRoomVersion,
                      MIncompatibleRoomVersion, MBadState, MGuestAccessForbidden, MCaptchaNeeded,
                      MCaptchaInvalid, MMissingParam, MInvalidParam, MTooLarge, MExclusive,
                      MUserDeactivated)
from .crypto import (CryptoError, EncryptionError, SessionShareError, DecryptionError,
                     MatchingSessionDecryptionError, DeviceValidationError, SessionNotFound,
                     DuplicateMessageIndex, VerificationError, DecryptedPayloadError,
                     MismatchingRoomError)
from .well_known import (WellKnownError, WellKnownNotJSON, WellKnownMissingHomeserver,
                         WellKnownNotURL, WellKnownUnexpectedStatus, WellKnownUnsupportedScheme,
                         WellKnownInvalidVersionsResponse)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Iterable, List, Any, TypeVar, Generic, Union, Callable, Dict

T = TypeVar('T')
Number = Union[int, float]


class Metric:
    name: str
    documentation: str
    unit: str
    typ: str
    samples: List[Any]

    def add_sample(self, name: str, labels: Iterable[str], value: Any, timestamp: Any = None,
                   exemplar: Any = None) -> None: ...


class MetricWrapperBase(Generic[T]):
    def __init__(self, name: str, documentation: str, labelnames: Iterable[str] = (),
                 namespace: str = "", subsystem: str = "", unit: str = "", registry: Any = None,
                 labelvalues: Any = None) -> None: ...

    def describe(self) -> List[Metric]: ...

    def collect(self) -> List[Metric]: ...

    def labels(self, *labelvalues, **labelkwargs) -> T: ...

    def remove(self, *labelvalues) -> None: ...


class ContextManager:
    def __enter__(self) -> None: ...

    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...

    def __call__(self, f) -> None: ...


class Counter(MetricWrapperBase['Counter']):
    def inc(self, amount: Number = 1) -> None: ...

    def count_exceptions(self, exception: Exception = Exception) -> ContextManager: ...


class Gauge(MetricWrapperBase['Gauge']):
    def inc(self, amount: Number = 1) -> None: ...

    def dec(self, amount: Number = 1) -> None: ...

    def set(self, value: Number = 1) -> None: ...

    def set_to_current_time(self) -> None: ...

    def track_inprogress(self) -> ContextManager: ...

    def time(self) -> ContextManager: ...

    def set_function(self, f: Callable[[], Number]) -> None: ...


class Summary(MetricWrapperBase['Summary']):
    def observe(self, amount: Number) -> None: ...

    def time(self) -> ContextManager: ...


class Histogram(MetricWrapperBase['Histogram']):
    def __init__(self, name: str, documentation: str, labelnames: Iterable[str] = (),
                 namespace: str = "", subsystem: str = "", unit: str = "", registry: Any = None,
                 labelvalues: Any = None, buckets: Iterable[Number] = ()) -> None: ...

    def observe(self, amount: Number = 1) -> None: ...

    def time(self) -> ContextManager: ...


class Info(MetricWrapperBase['Info']):
    def info(self, val: Dict[str, str]) -> None: ...


class Enum(MetricWrapperBase['Enum']):
    def __init__(self, name: str, documentation: str, labelnames: Iterable[str] = (),
                 namespace: str = "", subsystem: str = "", unit: str = "", registry: Any = None,
                 labelvalues: Any = None, states: Iterable[str] = None) -> None: ...

    def state(self, state: str) -> None: ...


def async_time(metric: Union[Gauge, Summary, Histogram]) -> Callable[[Callable], Callable]: ...



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Dict, Union, NewType
from enum import IntEnum, Enum
from attr import dataclass
import attr
import sys

from ..primitive import JSON, IdentityKey, SessionID, DeviceID
from ..util import SerializableAttrs, Serializable, ExtensibleEnum, Obj, deserializer
from .base import BaseRoomEvent, BaseUnsigned
from .message import RelatesTo


class EncryptionAlgorithm(ExtensibleEnum):
    OLM_V1: 'EncryptionAlgorithm' = "m.olm.v1.curve25519-aes-sha2"
    MEGOLM_V1: 'EncryptionAlgorithm' = "m.megolm.v1.aes-sha2"


class EncryptionKeyAlgorithm(ExtensibleEnum):
    CURVE25519: 'EncryptionKeyAlgorithm' = "curve25519"
    ED25519: 'EncryptionKeyAlgorithm' = "ed25519"
    SIGNED_CURVE25519: 'EncryptionKeyAlgorithm' = "signed_curve25519"


# IntEnum had a bug in Python 3.6 where it broke when using mixins
class OlmMsgType(Serializable, IntEnum if sys.version_info >= (3, 7) else Enum):
    PREKEY = 0
    MESSAGE = 1

    def serialize(self) -> JSON:
        return self.value

    @classmethod
    def deserialize(cls, raw: JSON) -> 'OlmMsgType':
        return cls(raw)


@dataclass
class OlmCiphertext(SerializableAttrs['OlmCiphertext']):
    body: str
    type: OlmMsgType


@dataclass
class EncryptedOlmEventContent(SerializableAttrs['EncryptedOlmEventContent']):
    ciphertext: Dict[str, OlmCiphertext]
    sender_key: IdentityKey
    algorithm: EncryptionAlgorithm = EncryptionAlgorithm.OLM_V1


@dataclass
class EncryptedMegolmEventContent(SerializableAttrs['EncryptedMegolmEventContent']):
    """The content of an m.room.encrypted event"""
    ciphertext: str
    sender_key: IdentityKey
    device_id: DeviceID
    session_id: SessionID
    _relates_to: Optional[RelatesTo] = attr.ib(default=None, metadata={"json": "m.relates_to"})
    algorithm: EncryptionAlgorithm = EncryptionAlgorithm.MEGOLM_V1

    @property
    def relates_to(self) -> RelatesTo:
        if self._relates_to is None:
            self._relates_to = RelatesTo()
        return self._relates_to

    @relates_to.setter
    def relates_to(self, relates_to: RelatesTo) -> None:
        self._relates_to = relates_to


EncryptedEventContent = NewType('EncryptedEventContent',
                                Union[EncryptedOlmEventContent, EncryptedMegolmEventContent])


@deserializer(EncryptedEventContent)
def deserialize_encrypted(data: JSON) -> Union[EncryptedEventContent, Obj]:
    alg = data.get("algorithm", None)
    if alg == EncryptionAlgorithm.MEGOLM_V1.value:
        return EncryptedMegolmEventContent.deserialize(data)
    elif alg == EncryptionAlgorithm.OLM_V1.value:
        return EncryptedOlmEventContent.deserialize(data)
    return Obj(**data)


setattr(EncryptedEventContent, "deserialize", deserialize_encrypted)


@dataclass
class EncryptedEvent(BaseRoomEvent, SerializableAttrs['EncryptedEvent']):
    """A m.room.encrypted event"""
    content: EncryptedEventContent
    _unsigned: Optional[BaseUnsigned] = attr.ib(default=None, metadata={"json": "unsigned"})

    @property
    def unsigned(self) -> BaseUnsigned:
        if not self._unsigned:
            self._unsigned = BaseUnsigned()
        return self._unsigned

    @unsigned.setter
    def unsigned(self, value: BaseUnsigned) -> None:
        self._unsigned = value



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional, Type
import json

from sqlalchemy import Column, Text, Boolean, types

from mautrix.types import (RoomID, PowerLevelStateEventContent as PowerLevels,
                           RoomEncryptionStateEventContent as EncryptionInfo, Serializable)
from mautrix.util.db import Base


class SerializableType(types.TypeDecorator):
    impl = types.Text

    def __init__(self, python_type: Type[Serializable], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._python_type = python_type

    @property
    def python_type(self) -> Type[Serializable]:
        return self._python_type

    def process_bind_param(self, value: Serializable, dialect) -> Optional[str]:
        if value is not None:
            return json.dumps(value.serialize())
        return None

    def process_result_value(self, value: str, dialect) -> Optional[Serializable]:
        if value is not None:
            return self.python_type.deserialize(json.loads(value))
        return None

    def process_literal_param(self, value, dialect):
        return value


class RoomState(Base):
    __tablename__ = "mx_room_state"

    room_id: RoomID = Column(Text, primary_key=True)
    is_encrypted: bool = Column(Boolean, nullable=True)
    has_full_member_list: bool = Column(Boolean, nullable=True)
    encryption: EncryptionInfo = Column(SerializableType(EncryptionInfo), nullable=True)
    power_levels: PowerLevels = Column(SerializableType(PowerLevels), nullable=True)

    @property
    def has_power_levels(self) -> bool:
        return bool(self.power_levels)

    @property
    def has_encryption_info(self) -> bool:
        return self.is_encrypted is not None

    @classmethod
    def get(cls, room_id: RoomID) -> Optional['RoomState']:
        return cls._select_one_or_none(cls.c.room_id == room_id)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, Union
from attr import dataclass
import attr

from ..primitive import JSON, RoomID, UserID
from ..util import SerializableAttrs, Obj, deserializer
from .base import EventType, BaseEvent


@dataclass
class RoomTagInfo(SerializableAttrs['RoomTagInfo']):
    order: Union[int, float, str] = None


@dataclass
class RoomTagAccountDataEventContent(SerializableAttrs['RoomTagAccountDataEventContent']):
    tags: Dict[str, RoomTagInfo] = attr.ib(default=None, metadata={"json": "tags"})


DirectAccountDataEventContent = Dict[UserID, List[RoomID]]


AccountDataEventContent = Union[RoomTagAccountDataEventContent, DirectAccountDataEventContent, Obj]
account_data_event_content_map = {
    EventType.TAG: RoomTagAccountDataEventContent,
}


# TODO remaining account data event types


@dataclass
class AccountDataEvent(BaseEvent, SerializableAttrs['AccountDataEvent']):
    content: AccountDataEventContent

    @classmethod
    def deserialize(cls, data: JSON) -> 'AccountDataEvent':
        try:
            data.get("content", {})["__mautrix_event_type"] = EventType.find(data.get("type"))
        except ValueError:
            return Obj(**data)
        return super().deserialize(data)

    @staticmethod
    @deserializer(AccountDataEventContent)
    def deserialize_content(data: JSON) -> AccountDataEventContent:
        evt_type = data.pop("__mautrix_event_type", None)
        content_type = account_data_event_content_map.get(evt_type, None)
        if not content_type:
            return Obj(**data)
        return content_type.deserialize(data)



# Copyright (c) 2020 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Type, Dict, Tuple, List, Any, TYPE_CHECKING

from mautrix.types import DeviceID, SyncToken, IdentityKey, SessionID, EventID, UserID, RoomID
from mautrix.util.file_store import FileStore

from .. import OlmAccount, Session, InboundGroupSession, OutboundGroupSession, DeviceIdentity
from .memory import MemoryCryptoStore

if TYPE_CHECKING:
    from typing import Protocol, TypedDict, TypeVar

    T = TypeVar('T')


    class Pickled(TypedDict, total=False):
        pickle: bytes


    class PickledStore(TypedDict):
        version: int
        device_id: DeviceID
        sync_token: SyncToken
        account: Pickled
        message_indices: Dict[Tuple[IdentityKey, SessionID, int], Tuple[EventID, int]]
        devices: Dict[UserID, Dict[DeviceID, DeviceIdentity]]
        olm_sessions: Dict[IdentityKey, List[Pickled]]
        inbound_sessions: Dict[Tuple[RoomID, IdentityKey, SessionID], Pickled]
        outbound_sessions: Dict[RoomID, Pickled]


    class Picklable(Protocol[T]):
        @classmethod
        def from_pickle(cls, pickle: bytes, passphrase: str, **kwargs: Any) -> T: ...

        def pickle(self, passphrase: str) -> bytes: ...


class PickleCryptoStore(FileStore, MemoryCryptoStore):
    path: str
    save_interval: float
    _last_save: float

    def __init__(self, account_id: str, pickle_key: str, path: str, save_interval: float = 60.0
                 ) -> None:
        FileStore.__init__(self, path=path, save_interval=save_interval, binary=True)
        MemoryCryptoStore.__init__(self, account_id, pickle_key)

    def deserialize(self, data: 'PickledStore') -> None:
        if data.get("version", 0) != 1:
            raise ValueError("Unsupported file crypto store version")
        self._device_id = data["device_id"]
        self._sync_token = data["sync_token"]
        self._account = self._from_pickle(OlmAccount, data["account"])
        self._message_indices = data["message_indices"]
        self._devices = data["devices"]
        self._olm_sessions = {identity_key: [self._from_pickle(Session, session)
                                             for session in sessions]
                              for identity_key, sessions in data["olm_sessions"].items()}
        self._inbound_sessions = {key: self._from_pickle(InboundGroupSession, session)
                                  for key, session in data["inbound_sessions"].items()}
        self._outbound_sessions = {room_id: self._from_pickle(OutboundGroupSession, session)
                                   for room_id, session in data["outbound_sessions"].items()}

    def _from_pickle(self, entity_type: Type['Picklable[T]'], data: 'Pickled') -> 'T':
        pickle_data = data.pop("pickle")
        return entity_type.from_pickle(pickle_data, passphrase=self.pickle_key, **data)

    def serialize(self) -> 'PickledStore':
        return {
            "version": 1,
            "device_id": self._device_id,
            "sync_token": self._sync_token,
            "account": {
                "pickle": self._account.pickle(self.pickle_key),
                "shared": self._account.shared
            },
            "message_indices": self._message_indices,
            "devices": self._devices,
            "olm_sessions": {identity_key: [{"pickle": session.pickle(self.pickle_key),
                                             "creation_time": session.creation_time,
                                             "use_time": session.use_time}
                                            for session in sessions]
                             for identity_key, sessions in self._olm_sessions.items()},
            "inbound_sessions": {key: {"pickle": session.pickle(self.pickle_key),
                                       "room_id": session.room_id,
                                       "sender_key": session.sender_key,
                                       "signing_key": session.signing_key,
                                       "forwarding_chain": session.forwarding_chain}
                                 for key, session in self._inbound_sessions.items()},
            "outbound_sessions": {room_id: {"pickle": session.pickle(self.pickle_key),
                                            "max_age": session.max_age,
                                            "max_messages": session.max_messages,
                                            "creation_time": session.creation_time,
                                            "use_time": session.use_time,
                                            "message_count": session.message_count,
                                            "room_id": session.room_id,
                                            "shared": session.shared}
                                  for room_id, session in self._outbound_sessions.items()}
        }

    async def put_device_id(self, device_id: DeviceID) -> None:
        await super().put_device_id(device_id)
        self._time_limited_flush()

    async def put_next_batch(self, next_batch: SyncToken) -> None:
        await super().put_next_batch(next_batch)
        self._time_limited_flush()

    async def put_account(self, account: OlmAccount) -> None:
        await super().put_account(account)
        self._save()

    async def add_session(self, key: IdentityKey, session: Session) -> None:
        await super().add_session(key, session)
        self._time_limited_flush()

    async def put_group_session(self, room_id: RoomID, sender_key: IdentityKey,
                                session_id: SessionID, session: InboundGroupSession) -> None:
        await super().put_group_session(room_id, sender_key, session_id, session)
        self._time_limited_flush()

    async def add_outbound_group_session(self, session: OutboundGroupSession) -> None:
        await super().add_outbound_group_session(session)
        self._time_limited_flush()

    async def update_outbound_group_session(self, session: OutboundGroupSession) -> None:
        await super().update_outbound_group_session(session)
        self._time_limited_flush()

    async def remove_outbound_group_session(self, room_id: RoomID) -> None:
        await super().remove_outbound_group_session(room_id)
        self._time_limited_flush()

    async def remove_outbound_group_sessions(self, rooms: List[RoomID]) -> None:
        await super().remove_outbound_group_sessions(rooms)
        self._time_limited_flush()

    async def put_devices(self, user_id: UserID, devices: Dict[DeviceID, DeviceIdentity]) -> None:
        await super().put_devices(user_id, devices)
        self._time_limited_flush()
